%
% get constraints in CIF, map them to RDF
% and send them over to the AKT-KRAFT gateway
%
% This shows that CIF constraints can be mapped to RDF,
% send as AKT messages and converted back to Prolog terms
% and then use in KRAFT reasoning
%

:- [sample_constraint].

:- ['~/quail_local/aktbus/prolog/src/absl'].	%load AKT Bus module
:- ['../../prolog/rdf2triple/rdf2triple'].	%load the rdf2triple module
:- ['../../prolog/util/print_xml'].

:- ['../fusing_mediator/transform'].
:- ['../../prolog/util/xcif2cif'].
:- ['../../prolog/util/print_cif'].
:- ['../../prolog/util/cif2xml'].
:- use_module(library(lists)).

:- [url].		%load sender & receiver URL

go :-
	menu(Cs),				%get list of constraints to send
	preprocess(Cs,Cif),			%map them to CIF
	compose_aktmessage(Cif,Aktmessage),	%compose constraints into AKT message
	print_xml:print_xml(Aktmessage),
	nl,
	send_akt_message(Aktmessage).

/*
test([]).
test([X|Xs]) :-
	Params=[representation=classic,domain_rdfs='http://www.csd.abdn.ac.uk/~khui/akt/ix/pcconfig'],
	cif2xml:cif2xml(Params,C,X),
	test(Xs).
*/
menu(Cs) :-
	write('*** LIST OF CONSTRAINTS'),
	nl,
	print_constraints,						%print list of sample constraints
	write('*** Please select constraints to send (as a list). Or type "receive" to get any reply:'),
	read(Cs).							%read selection list

%
% print list of known constraints
%
print_constraints :-
	sample_constraint:constraint(C),
	sample_constraint:detail(C,description,Desc),
	write(C),
	write(':'),
	write(Desc),
	nl,
	fail.
print_constraints.


%
% map I-code constraints to CIF form
%
preprocess([],[]).
preprocess([C|Cs],[X|Xs]) :-
	sample_constraint:detail(C,cif,Cif),
	transform:xcif(Cif,Xcif),		%map it to XCIF
	xcif2cif:xcif2cif(Xcif,T),		%map it back to CIF w/ missing quantification filled
	Params=[representation=classic,domain_rdfs='http://www.csd.abdn.ac.uk/~khui/akt/ix/pcconfig'],
	cif2xml:cif2xml(Params,T,X),		%map it to XML
	write('==='),nl,
	print_xml:print_xml(X),
	nl,
	write('==='),nl,
	preprocess(Cs,Xs),
	!.

%
% compose an AKT message
%
compose_aktmessage(Cif,Aktmessage) :-
	write('***composing AKT message'),
	nl,
%	cif_into_content(Cif,Content,Cif_namespace),
	extract_rdf_constraints(Cif,Rdf_constraints,Cif_namespace),		%get constraints in RDF, removing XML header
	sender_url(Sender),
	receiver_url(Receiver),
	compose_kqml_message(Sender,Receiver,'ask_one',Rdf_constraints,Cif_namespace,Aktmessage),
%	absl:compose_message(Sender,Receiver,Content,[Cif_namespace],Aktmessage),
	!.

compose_aktmessage(_,_) :-
	write('*** FAIL composing AKT message'),
	nl,
	fail.

extract_rdf_constraints(Cifs,Constraints,'xmlns:cif'=Url) :-
	Goal=(	member(Cif,Cifs),			%get a constraint from list
		Cif=[_,X_rdf],				%forget about the XML header
		X_rdf=env('rdf:RDF',_,X_constraint)	%take the RDF subtree & get the constraint
		),
	Template=X_constraint,
	findall(Template,call(Goal),Constraints),	%get all constraints in XML tree
	%
	% find namespace of CIF here
	%
	Cifs=[First_cif|_],				%get a CIF
	First_cif=[_,Rdf_tree],
	Rdf_tree=env('rdf:RDF',Attribs,_),
	member('xmlns:cif'=Url,Attribs).
%	Content=env('aktbus:constraint_content',[],Constraints).

%
% compose an KQML message
%
compose_kqml_message(Sender,Receiver,Perf,Rdf_constraints,Cif_namespace,Message) :-
	Kqml_url='http://www.csd.abdn.ac.uk/~khui/akt/akt-kqml#',
	Kqml_content=env('akt-kqml:KqmlContent',['xmlns:akt-kqml'=Kqml_url,Cif_namespace],Kqml_params),
	Kqml_params=[	Has_performative,
			Has_language,
			Has_ontology,
			Has_sender,
			Has_receiver,
			Has_reply_with,
			Has_content
			],
	Reply_with=cs_100,
	Has_performative=env('akt-kqml:hasPerformative',[],Perf),
	Has_language=env('akt-kqml:language',[],cif),
	Has_ontology=env('akt-kqml:ontology',[],shared),
	Has_sender=env('akt-kqml:sender',[],Sender),
	Has_receiver=env('akt-kqml:receiver',[],Receiver),
	Has_reply_with=env('akt-kqml:reply_with',[],Reply_with),
	Has_content=env('akt-kqml:content',[],Rdf_constraints),
	absl:compose_message(Sender,Receiver,Kqml_content,[],Message).



send_akt_message(Message) :-
	write('***sending AKT message'),
	nl,
	absl:send_message(Message),
	!.

send_akt_message(_) :-
	write('*** FAIL sending AKT message'),
	nl,
	fail.


/*

go :-
	receiver_url(Url),
	go(Url).

go(Receiver) :-
	rdf2triple:clean_triple,			%clean up all triples in module rdf2triple
	add_all_triples,				%assert triples into module rdf2triple
	rdf2triple:triple2rdf(abbreviated,id(ip1),Rdf),	%map triples to RDF tree
	Rdf=env('rdf:RDF',_,Content),
	%
	write('*** sending issue:'),
	nl,
	Sender='http://quail.csd.abdn.ac.uk/~khui/akt/makeup_panel',
	Extra_namespace=['xmlns:ix_issue'='http://www.csd.abdn.ac.uk/~khui/akt/ix/issue-rdfs.xml'],
	absl:compose_message(Sender,Receiver,Content,Extra_namespace,Message),
	print_xml:print_xml(Message),
	nl,
	absl:send_message(Message).

add_all_triples :-
	namespace(X,Y),
	assertz(rdf2triple:namespace(X,Y)),
	fail.

add_all_triples :-
	triple(X,Y,Z),
	assertz(rdf2triple:triple(X,Y,Z)),
	fail.

add_all_triples.


%
% the issue defined as triples
%
namespace(rdf, 'http://www.w3.org/1999/02/22-rdf-syntax-ns#').
namespace(rdfs, 'http://www.w3.org/2000/01/rdf-schema#').
%namespace(cif, 'http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-compact-rdfs.xml#')
namespace(ix_issue,'http://www.csd.abdn.ac.uk/~khui/akt/ix/issue-rdfs.xml#').

triple(id(ip1),rdf:type,ix_issue:issue_packet).

triple(id(ip1),ix_issue:issue_packet_issue_name,'solve_technical_constraint').
triple(id(ip1),ix_issue:issue_packet_receiver,'aberdeen').
triple(id(ip1),ix_issue:issue_packet_receiver_role,'technical').
triple(id(ip1),ix_issue:issue_packet_sender,'edinburgh').
triple(id(ip1),ix_issue:issue_packet_sender_role,'marketing').
triple(id(ip1),ix_issue:issue_packet_instance,anonymous(i)).

triple(anonymous(i),rdf:type,ix_issue:issue).
triple(anonymous(i),ix_issue:issue_status,created).
triple(anonymous(i),ix_issue:issue_priority,high).
triple(anonymous(i),ix_issue:issue_commmunicator,anonymous(c)).
triple(anonymous(i),ix_issue:issue_begin_time,anonymous(bt)).
triple(anonymous(i),ix_issue:issue_end_time,anonymous(et)).
triple(anonymous(i),ix_issue:issue_content,anonymous(sol)).

triple(anonymous(bt),rdf:type,ix_issue:time).
triple(anonymous(bt),ix_issue:time_year,'2002').
triple(anonymous(bt),ix_issue:time_month,'10').
triple(anonymous(bt),ix_issue:time_day,'21').
triple(anonymous(bt),ix_issue:time_hour,'18').
triple(anonymous(bt),ix_issue:time_minute,'44').
triple(anonymous(bt),ix_issue:time_second,'0').

triple(anonymous(et),rdf:type,ix_issue:time).
triple(anonymous(et),ix_issue:time_year,'2002').
triple(anonymous(bt),ix_issue:time_month,'10').
triple(anonymous(et),ix_issue:time_day,'21').
triple(anonymous(et),ix_issue:time_hour,'18').
triple(anonymous(et),ix_issue:time_minute,'45').
triple(anonymous(et),ix_issue:time_second,'30').

triple(anonymous(c),rdf:type,ix_issue:communicator).
triple(anonymous(c),ix_issue:communicator_requestor,edinburgh).
triple(anonymous(c),ix_issue:communicator_requestor_type,marketing).
triple(anonymous(c),ix_issue:communicator_provider,aberdeen).
triple(anonymous(c),ix_issue:communicator_provider_type,technical).

triple(anonymous(sol),rdf:type,ix_issue:solution).
triple(anonymous(sol),ix_issue:solution_customer,john).
triple(anonymous(sol),ix_issue:solution_requirement,'constraint(a,b)').
%triple(anonymous(sol),ix_issue:solution_old_solution,).
%triple(anonymous(sol),ix_issue:solution_new_solution,).
%triple(anonymous(sol),ix_issue:solution_difference,).
*/

