
% supporting predicates for ECLiPSe used by the ECLiPSe solver module
%
/*
:- module(eclipse_stdlib,[	find_prolog_var/3,		%bind Daplex vars to ECLiPSe vars
				declare_vars/2,			%given a list of Daplex:Type, create a list of Daplex:Prolog
				label_vars/1,			%label all in a list
				exist/3,			%the existential quantifier
				forall/3,			%the universal quantifier
				all_violate/3,			%the universal quantifier by negation
				if_then_else/5			%conditional constraint
			]).
*/
%:- use_module(library(suspend)).
%:- use_module(library(fd)).
%:- use_module(library(propia)).
%:- use_module('common/util/is_daplex_variable').

:- ['../../prolog/util/retractall'].

:- dynamic(debug_msg/2).		%debug message flag


%
% find_prolog_var/3
%
% given a (list of) Daplex var, find the corresponding Prolog var
%
% find_prolog_var(+Daplex,+Var_pairs,-Prolog)
%
% Daplex	=	Daplex var to find
% Var_pairs	=	list of Daplex:Prolog pairs
% Prolog	=	the Prolog var in ECLiPSe
%

find_prolog_var(Daplex,Var_pairs,Prolog) :-
	member(Daplex:Prolog,Var_pairs),
	!.

find_prolog_var([],_,[]).

find_prolog_var([D|Ds],Var_pairs,[P|Ps]) :-
	is_daplex_variable:is_daplex_variable(D),
	find_prolog_var(D,Var_pairs,P),
	find_prolog_var(Ds,Var_pairs,Ps).


%
% declare_vars/2
%
% given a list of Daplex:Type, declare finite domain vars
% and returns a list of Daplex:Prolog_var with domains initialised
%
% declare_vars(+Var_with_type,-Var_with_prolog)
%
% Var_with_type		=	a Daplex:Type pair
%				OR a list of Daplex:Type pairs
% Var_with_prolog	=	a Daplex:Prolog pair
%				OR a list of Daplex:Prolog pairs
%				with all Prolog (ECLiPSe vars) initialised
%				with proper domains
%

declare_vars([],[]).

declare_vars(Daplex:Type,Daplex:Prolog) :-
	set_domain(Daplex,Type,Prolog).		%set its domain

declare_vars([V|Vs],[D|Ds]) :-
	declare_vars(V,D),
	declare_vars(Vs,Ds).



%
% label_vars/1
%
% label all variables
%
% label_vars(?Var_list)
%
% Var_list	=	a list in the form daplex:Prolog to be labelled
%
label_vars([]).
label_vars([_:X|Tail]) :-
	indomain(X),
	label_vars(Tail).

label([]).
label([H|T]) :-
	indomain(H),
	label(T).


%
% set_domain/3
%
% set the domain of Daplex to the entity class Class
%
% set_domain(+Daplex,+Class,-Prolog)
%
% Daplex	=	a daplex var
% Class		=	an entity class
% Prolog	=	the ECLiPSe FD var
%

set_domain(_,integer,Prolog) :-
	!,
	Prolog :: 1..32767.
set_domain(_,string,_) :- !.
set_domain(_,boolean,Prolog) :-
	!,
	Prolog::[true,false].
set_domain(_,float,_) :- !.
set_domain(_,Class,Prolog) :-
	object(Class,Prolog) infers most.



%
% USER-DEFINED CONSTRAINT PREDICATES
%

%
% exist/3
%
% the existential quantifier
%
% exist(?Vars,+Gen,+Pred)
%
% Vars	=	a list of ECLiPSe vars
% Gen	=	the generator as ECLiPSe codes
% Pred	=	the predicate as ECLiPSe codes
%

exist(_,G,P) :-
	call(G),
	call(P).





%
% if_then_else/5
%
% A conditional constraint, by default using the ACTIVE version.
%
% if_then_else(+Flag,?Vars,?Cond,?Then,?Else)
%
% Flag	=	either 'active/passive'
% Vars	=	the list of involved vars
% Cond	=	the testing condition
%		NOTE:	For the active version, please use the
%			"constraint"/ECLiPSe version for the test.
%			For the passive version, plesae use the
%			standard Prolog code.
% Then	=	the goal to call when Cond succeeds
% Else	=	the goal to call when Cond fails
%

%
% testing predicate
%
ite_if_then_else(active,V,Cond,Then,Else) :-
	ite_check_condition(V,Cond,Result),
	ite_if_then_else_part2(Result,V,Cond,Then,Else).


ite_if_then_else_part2(true,_,_,Then,_) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: call THEN'),
	call(Then),
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: THEN called').

%
% if condition must fail, call ELSE
%
ite_if_then_else_part2(fail,_,_,_,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: call ELSE'),
	call(Else).

%
% if not sure, suspend it until vars are further constrained
%
ite_if_then_else_part2(not_sure,V,Cond,Then,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: suspending'),
	suspend(ite_if_then_else(active,V,Cond,Then,Else),8,->(V,constrained)).

ite_check_condition(V,Cond,true) :-
	ite_condition_must_succeed(V,Cond),	%if cond must succeed, return true
	!.					%cut to prevent backtracking into another clause

ite_check_condition(V,Cond,fail) :-
	condition_must_fail(V,Cond),		%if condition MUST fail
	!.					%then CUT it to prevent backtracking into the 2nd clause

ite_check_condition(_,_,not_sure).

ite_condition_must_succeed(Vs,Cond):-
	member(V,Vs),
	delete(V,Vs,Vs_left),
	indomain(V),
	get_domains(Vs_left,Domains),
	copy_domains(Domains,Before),
	subcall(Cond,[]),
	get_domains(Vs_left,After),
	Before\=After,
	!,
	fail.

ite_condition_must_succeed(_,Cond) :-
	subcall(Cond,[]).











%
% if_then_else/5
%
% the PASSIVE version
%
if_then_else(passive,V,Cond,Then,Else) :-
	nonground(V),								%if not all vars are grounded
	!,
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: vars non-grounded, suspending'),
	suspend(if_then_else(passive,V,Cond,Then,Else),8,->(V,inst)).	%suspend until some vars are instantiated
%	suspend(passive_if_then_else_decision(Cond,Then,Else),8,->(V,inst)).	%suspend until they're all instantiated


if_then_else(passive,_,Cond,Then,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: all vars grounded, making decision'),
	passive_if_then_else_decision(Cond,Then,Else).				%make decision

/*
*
* old but working version
*
if_then_else(passive,V,Cond,Then,Else) :-
	nonground(V),								%if not all vars are grounded
	!,
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: vars non-grounded, suspending'),
	suspend(passive_if_then_else_decision(Cond,Then,Else),8,->(V,inst)).	%suspend until they're all instantiated


if_then_else(passive,_,Cond,Then,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: all vars grounded, making decision'),
	passive_if_then_else_decision(Cond,Then,Else).				%make decision
*/

%
% if_then_else/5
%
% the ACTIVE version
%
if_then_else(active,V,Cond,Then,Else) :-
	check_condition(V,Cond,Result),				%check the testing condition
	active_if_then_else_part2(Result,V,Cond,Then,Else).	%act according to test result


%
% passive if_then_else wake up & make decision
%
passive_if_then_else_decision(Cond,Then,_) :-
	call(Cond),				%call Cond
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: call THEN'),
	call(Then).				%if succeeds, call THEN

passive_if_then_else_decision(Cond,_,Else) :-
	~(call(Cond)),				%sounded negation of Cond
	print_debug_message(if_then_else,'***IF-THEN-ELSE/passive: call ELSE'),
	call(Else).				%call ELSE


%
% active if_then_else make decision
%
% if condition must succeed, call THEN
%
active_if_then_else_part2(true,_,_,Then,_) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: call THEN'),
	call(Then),
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: THEN called').

%
% if condition must fail, call ELSE
%
active_if_then_else_part2(fail,_,_,_,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: call ELSE'),
	call(Else).

%
% if not sure, suspend it until vars are further constrained
%
active_if_then_else_part2(not_sure,V,Cond,Then,Else) :-
	print_debug_message(if_then_else,'***IF-THEN-ELSE/active: suspending'),
	suspend(if_then_else(active,V,Cond,Then,Else),8,->(V,constrained)).

%
% check_condition/3
%
% check whether a condition must succeed/fail for active if-then-else
%
% To test whether a condition must be true without instantiating any var
% we compare the var domains before & after applying the condition as a constraint.
% If there is no delayed goal and the domains do not change
% then we know that applying the testing condition (as a constraint) do not
% remove any value. Thus it must be true!
%
% NOTE: We don't mind the testing condition is applied on the vars as
%	it doesn't change the domains at all.
%
check_condition(V,Cond,true) :-
	condition_must_succeed(V,Cond),		%if cond must succeed, return true
	!.					%cut to prevent backtracking into another clause

%
% ok, now we have only 2 possibilities
% 1)	Either the constraint (of the test) cannot be applied, and it failed.
% 2)	Or it can apply, but there are delayed goal
%	OR the domains changed.
%
check_condition(V,Cond,fail) :-
	condition_must_fail(V,Cond),		%if condition MUST fail
	!.					%then CUT it to prevent backtracking into the 2nd clause

%
% now we're not sure if the test succeeds/fails
%
check_condition(_,_,not_sure).			%return 'not_sure'




%
% test whether condition MUST succeed
%
% NOTE:	We don't mind if Cond is applied on vars
%	as it doesn't change the domains and
%	no goal is delayed in calling it
%
/*
condition_must_succeed(Vs,Cond) :-
	get_domains(Vs,Domains),			%get domains before applying test (in constraint form)
	copy_domains(Domains,Before),		%make a copy of all domains
	subcall(Cond,[]),			%imposing the the condition (as a constraint) does not delay anything
	member(V,Vs),				%choose a var
	indomain(V),				%label it
	get_domains(V,After),			%get domains after applying
	different(Before,After),

*/

condition_must_succeed(V,Cond) :-
	get_domains(V,Domains),			%get domains before applying test (in constraint form)
	copy_domains(Domains,Before),		%make a copy of all domains
	subcall(Cond,[]),			%imposing the the condition (as a constraint) does not delay anything
	get_domains(V,After),			%get domains after applying
	Before=After.				%and domains do not change


%
% we try to apply the test (as a constraint),
% if it DOES NOT fail, then there must be delayed goal
% OR the domains changed.
% In this case, we have to undo the application of constraint
% (thus the FAIL) and return 'not_sure'.
%
condition_must_fail(_,Cond) :-
	call(Cond),			%try to apply the test (as a constraint)
	!,
	fail.				%if it can be apply, FAIL it, 'cause we have to remove its effect
					%the CUT prevents backtracking on check_condition_part2/3
					%thus it will backtrack on the calling predicate

condition_must_fail(_,_).





%
% forall/3
%
% the universal quantifier
%
% forall(?Vars,+Gen,+Pred)
%
% Vars	=	the list of involved variables
% Gen	=	the ECLiPSe code to define the initial domains
% Pred	=	the predicate to enforce
%

forall(V,G,P) :-
	%
	% we have to make a copy of all vars for us to work on
	%
	copy_term((V,G),(V_copy,G_copy)),			%make a copy of vars & G
	call(G),						%apply the generator G to initialise domains
	call(G_copy),						%apply the generator for V_copy
	print_debug_message(forall,'***FORALL: initial domains created'),
	%
	% before we proceed, we have to find the set vars
	% which are related by their generators
	%
	group_related_vars(G_copy,Related_var_groups),		%group related vars
	print_debug_message(forall,['***FORALL: original vars:',V_copy]),
	print_debug_message(forall,['***FORALL: groups of related vars:',Related_var_groups]),
	%
	% now initialise the bad domains and proceed
	%
	make_initial_bad_domains(V,Bad),			%initially there is nothing bad we know
	forall_main(V,P,V_copy,Bad,Related_var_groups).		%now go to the main


%
% forall_main/5
%
% the main predicate for remove bad values in a universal quantifier
%
forall_main(V,P,V_copy,Bad,Related_var_groups) :-
	print_debug_message(forall,'***FORALL: try applying P'),
	%
	% try applying P and see what happens. There are only 2 cases:
	% 1)	it applies with NO delayed-goal
	% 2)	it applies WITH delayed-goal
	%
	% if P doesn't apply, try_apply_p/4 simply fails & forall/3 fails
	% *** NOTE: try_applying_p/4 may backtrack
	%
	try_applying_p(V,P,Delayed_flag,Removed),
	print_debug_message(forall,['***FORALL: Result & Removed instance:',Delayed_flag,Removed]),
	forall_part2(Delayed_flag,V,P,V_copy,Bad,Removed,Related_var_groups).		%proceeds

%
% forall_part2/7
%
% take action depending on the result of "trying to" apply P
%
% forall_part2(+Result,?Vars,+Pred,?Vars_copy,+Bad_domains,?Removed,?Related_var_groups)
%
% Result	=	no_delayed_goal/has_delayed_goal
%			"no_delayed_goal" means P applies without delayed-goal in all cases
%			"has_delayed_goal" means P applies with delayed-goal in some cases
% Vars		=	the list of involved variables
% Pred		=	the predicate to enforce
% Vars_copy	=	a copy of all involved variables
% Bad_domains	=	known bad values in all domains
% Removed	=	if Result=no_delayed_goal, then this is the list of values removed when P applies
%			if Result=has_delayed_goal, this is the list of values that are always removed
% Related_var_groups	=	list of related FD vars in the form [group1,group2,...]
%				where each group is a list of FD vars
%				This tells that vars in a group are related by the generator
%


%
% if P applies without delayed-goal, then all values in domains are valid
%
forall_part2(no_delayed_goal,V,_,V_copy,Bad,Removed,Related_var_groups) :-
	print_debug_message(forall,'***FORALL: P applies without delayed-goal'),
	get_domain_union(Bad,Removed,New_bad),				%combine bad values
	remove_bad_values(V,V_copy,New_bad,Related_var_groups).		%remove bad values from V's domains

%
% if P applies with delayed-goals, then come back when vars are further constrained
%
forall_part2(has_delayed_goal,V,P,V_copy,Bad,Removed,Related_var_groups) :-
	print_debug_message(forall,'***FORALL: P applies with delayed-goal'),
	get_domain_union(Bad,Removed,New_bad),							%get all bad values
	remove_bad_values(V,V_copy,New_bad,Related_var_groups),					%remove bad values
	print_debug_message(forall,'***FORALL: suspending'),
	suspend(forall_main(V,P,V_copy,New_bad,Related_var_groups),7,->(V,constrained)).	%suspend until fruther constrained

%
% make the initial list of empty bad domains
%
make_initial_bad_domains([],[]).
make_initial_bad_domains([_|Ds],[[]|Bads]) :-
	make_initial_bad_domains(Ds,Bads).


%
% try_applying_p/4
%
% try applying predicate P (as a constraint) and look for any change in domains,
% collect any value being removed.
%
try_applying_p(V,P,Delayed_flag_instance,Removed_instance) :-
	get_domains(V,Domains),						%get domains before applying P
	copy_domains(Domains,Before),					%make a copy
	%
	% get domain differences for all P application instances
	%
	Goal=(	subcall(P,Delayed),					%call P & find delayed-goals
		map_delay_status(Delayed,Delayed_flag),			%map delayed-goal list into a flag
		get_domains(V,After),					%get domains after applying P
		get_domain_difference(Before,After,Removed)		%find the difference
	),
	Template=(Delayed_flag,Removed),				%the result template
	findall(Template,Goal,All_instances),
	%
	% get an application instance
	% ***NOTE:	it may backtrack
	%		and if P doesn't apply, member/2 simply fails & forall/3 fails
	%
	member((Delayed_flag_instance,Removed_instance),All_instances).	%get an instance ***NOTE: may backtrack


%
% map_delay_status/2
%
% check to see if there is any delayed-goal
%
% Because of a "strange" limitation of ECLiPSe, findall/3 cannot
% return the list of delayed-goal. Instead, we have to map it
% into a ok/delayed_goal for later reference.
%
map_delay_status([],no_delayed_goal).
map_delay_status([_|_],has_delayed_goal).



%
% end of var list
%
remove_bad_values([],[],_,_).

%
% for every var, check if there is any value that causes other vars to take bad values
%
remove_bad_values([V|Vs],[V_copy|Vs_copy],[Bad|Bads],Related_var_groups) :-
	print_debug_message(forall,'***FORALL: removing bad values'),
	%
	% remove bad values from V_copy's domain
	%
	dvar_domain(V_copy,V_copy_dom),					%find V_copy's current domain
	(	(	Bad=[],						%if no bad value for V_copy
			!,
			V_copy_temp_valid_dom=V_copy_dom		%then just keep it unchange
		);
		dom_difference(V_copy_dom,Bad,V_copy_temp_valid_dom,_)	%otherwise remove the bad values
	),
	%
	% now V_copy_temp_valid_dom contains no bad value of V
	%
%	dvar_update(V_copy,V_copy_temp_valid_dom),			%update V_copy's domain
	print_debug_message(forall,['***FORALL: after removing surely bad:',V_copy_temp_valid_dom]),
	%
	% after removing bad-values in V_copy
	% now look for any value in V_copy that will cause others to take bad values
	%
	% 1st, we have to find the list of related vars in order
	%
	print_debug_message(forall,['***FORALL: all related var groups',Related_var_groups]),
	(	(member(Group,Related_var_groups),		%find a group in the list
		var_member(V_copy,Group),			%where V_copy is a member
%		write('V_copy:'),write(V_copy),nl,
%		write('Vs_copy:'),write(Vs_copy),nl,
%		write('Group:'),write(Group),nl,
		whatever(Vs_copy,Group,Vars_related),
		print_debug_message(forall,['***FORALL: related vars:',Vars_related]),
		!
		);
		Vars_related=[]					%if V_copy is not in any group, then no one is related to it
	),
	%
	% now we proceed to remove bad values
	%
	print_debug_message(forall,'***FORALL: looking for bad values in current var'),

	Goal=(Vars_related,Bads)^
			(	indomain(V_copy),			%label the copy of V
				has_bad_value(Vars_related,Bads)		%and some vars in Vs_copy got bad values
			),
	(	(	setof(V_copy,Goal,Bad_list_for_v_copy),		%find the set of bad values of V_copy
			!
		);
		Bad_list_for_v_copy=[]					%or when no bad value can be found
	),
%	print_debug_message(forall,['***FORALL: indirect bad values:',Bad_list_for_v_copy]),
	%
	% remove these bad values from V
	%
	(	(	%
			% if there is no bad value of V_copy found,
			% just leave V untouched
			%
			Bad_list_for_v_copy=[],
			!,
			print_debug_message(forall,'***FORALL: all values of V_copy are OK'),
			dvar_domain(V,V_current_domain),
			dom_intersection(V_current_domain,V_copy_temp_valid_dom,Valid_dom_for_v,_),
			dvar_update(V,Valid_dom_for_v)
		);
		(	%
			% bad values found
			% remove them from V's domain
			%
			sorted_list_to_dom(Bad_list_for_v_copy,Bad_dom_for_v_copy),			%convert the list into a domain
			print_debug_message(forall,['***FORALL: indirect bad values:',Bad_dom_for_v_copy]),
			dom_difference(V_copy_temp_valid_dom,Bad_dom_for_v_copy,Valid_dom_for_v_copy,_),
			dvar_update(V_copy,Valid_dom_for_v_copy),
			dvar_domain(V,V_current_domain),						%get V's current domain
			dom_intersection(V_current_domain,Valid_dom_for_v_copy,Valid_dom_for_v,_),	%get the intersection
			dvar_update(V,Valid_dom_for_v)							%update V's domain
		)
	),
	remove_bad_values(Vs,Vs_copy,Bads,Related_var_groups).

%
% had_bad_value/2
%
% give a list of vars,
% check whether anyone of them can take a bad value
%

%
% the bad domain is empty, no one can take any bad value
%
has_bad_value([_|Vs],[[]|Bads]) :-
	!,
	has_bad_value(Vs,Bads).		%proceed to test the rest

has_bad_value([V|_],[Bad|_]) :-
	dvar_domain(V,D),		%get the domain of V
	dom_intersection(D,Bad,_,_),	%if it can intersect the bad domain,
	!.				%then is may take a bad value

has_bad_value([_|Vs],[_|Bads]) :-
	has_bad_value(Vs,Bads).		%or some var in the tail can take bad value



%
% copy_domains/2
%
% copy a list of domains
%
copy_domains([],[]).
copy_domains([D|Ds],[C|Cs]) :-
	dom_copy(D,C),
	copy_domains(Ds,Cs).

%
% get_domains/2
%
% get the value domains of a list of vars
%
% get_domains(?Vars,+Domains)
%
% Vars		=	a list of ECLiPSe FD vars
% Domains	=	a list of domains
%

get_domains([],[]).					%end of var list

get_domains([V|Vs],[D|Ds]) :-
	dvar_domain(V,D),
	get_domains(Vs,Ds).


%
% get_domain_difference/3
%
% given 2 lists of domains, compute their difference
%
% get_domain_difference(+List1,+List2,+Difference)
%
% List1		=	the 1st list of domains
% List2		=	the 2nd list of domains
% Difference	=	a list of differences between corresponding domains
%

get_domain_difference([],[],[]).				%end of list

get_domain_difference([A|As],[[]|Bs],[A|Ds]) :-			%when 2nd domain is empty
	!,
	get_domain_difference(As,Bs,Ds).			%recursively work on the rest

get_domain_difference([[]|As],[_|Bs],[[]|Ds]) :-		%when 1st domain is empty
	!,
	get_domain_difference(As,Bs,Ds).			%recursively work on the rest

get_domain_difference([X|As],[X|Bs],[[]|Ds]) :-		%when 1st domain is empty
	!,
	get_domain_difference(As,Bs,Ds).			%recursively work on the rest

get_domain_difference([A|As],[B|Bs],[D|Ds]) :-
	dom_difference(A,B,D,_),				%compute the difference of 2 domains
	get_domain_difference(As,Bs,Ds).			%recursively work on the rest


%
% get_domain_intersection/3
%
% given2 lists of domains, get their intersections
%
% get_domain_intersection(+List1,+List2,+Intersection)
%
% List1		=	the 1st list of domains
% List2		=	the 2nd list of domains
% Intersection	=	a list of intersection between corresponding domains
%

get_domain_intersection([],[],[]).

get_domain_intersection([[]|As],[_|Bs],[[]|Cs]) :-
	!,
	get_domain_intersection(As,Bs,Cs).

get_domain_intersection([_|As],[[]|Bs],[[]|Cs]) :-
	!,
	get_domain_intersection(As,Bs,Cs).

get_domain_intersection([A|As],[B|Bs],[C|Cs]) :-
	dom_intersection(A,B,C,_),
	!,
	get_domain_intersection(As,Bs,Cs).

get_domain_intersection([_|As],[_|Bs],[[]|Cs]) :-
	get_domain_intersection(As,Bs,Cs).


%
% get_domain_union/3
%
% combine 2 list of domains together
%
%
%
%

get_domain_union([],[],[]).			%end of list

get_domain_union([[]|As],[B|Bs],[B|Cs]) :-
	!,
	get_domain_union(As,Bs,Cs).

get_domain_union([A|As],[[]|Bs],[A|Cs]) :-
	!,
	get_domain_union(As,Bs,Cs).

get_domain_union([A|As],[B|Bs],[C|Cs]) :-
	dom_union(A,B,C,_),
	get_domain_union(As,Bs,Cs).



%
% all_violate/3
%
% a universal quantifier by negation
%
% all_violate(?Vars,+Gen,+Pred)
%
% Vars		=	a list of ECLiPSe vars
% Gen		=	the generator in ECLiPSe codes
% Pred		=	the negated version of the predicate in ECLiPSe codes
%

all_violate(V,G,P) :-
	not_exist(V,G,P).

%
% there exist no value in variables generated by G
% such that P is satisfied
%
not_exist(V,G,P) :-
	copy_term((V,G,P),(V_copy,G_copy,P_copy)),
	call(G_copy),
	call(P_copy),
	!,
	findall(V_copy,label(V_copy),Violating_v),
	call(G),
	remove_values(Violating_v,V).

not_exist(_,G,_) :-
	call(G).

remove_values([],_).
remove_values([V|Vs],Vars) :-
	remove_values2(V,Vars),
	remove_values(Vs,Vars).

remove_values2([],_).
remove_values2([V|Vs],[X|Xs]) :-
	X #\= V,
	remove_values2(Vs,Xs).








%
% set_debug/2
%
% setting the debug message flag
%

set_debug(Flag,Status) :-
	retractall(debug_msg(Flag,_)),
	asserta(debug_msg(Flag,Status)).


%
% print a debug message
%
print_debug_message(Flag,[]) :-
	debug_msg(Flag,on),
	!,
	nl.

print_debug_message(Flag,[H|T]) :-
	debug_msg(Flag,on),
	!,
	write(H),
	print_debug_message(Flag,T).

print_debug_message(Flag,Message) :-
	debug_msg(Flag,on),
	!,
	write(Message),
	nl.

print_debug_message(_,_).



%
% given a generator, find groups of related vars
%
% we focus on fnval/5 & find_prolog_var/3
%
group_related_vars(Goal,Var_groups) :-
	group_related_vars(Goal,[],Var_groups).		%start w/ an empty list

%
% cater for a list
%
group_related_vars([],Groups,Groups) :-
	!.						%end of list, no change in groups

group_related_vars([G|Gs],Old_groups,Var_groups) :-
	!,
	group_related_vars(G,Old_groups,Temp_groups),	%process the 1st one
	group_related_vars(Gs,Temp_groups,Var_groups).	%and the rest of the list

%
% conjoined goals
%
group_related_vars(Goal,Old_groups,Var_groups) :-
	Goal=(Subgoal1,Subgoal2),
	!,
	group_related_vars(Subgoal1,Old_groups,Temp_groups),
	group_related_vars(Subgoal2,Temp_groups,Var_groups).

%
% fnval/5 infers most
%
group_related_vars(Goal,Old_groups,Var_groups) :-
	Goal=(fnval(_,_,Args,_,Result) infers _),			%get args & result vars
	!,
	add_var_into_group([Result|Args],Old_groups,Var_groups).	%add them into list of groups

/*
group_related_vars(Goal,Groups,Groups) :-
	Ommit=[	find_prolog_var(_,_,_),
		set_domain(_,_,_),
		],
	member(Goal,Ommit),
	!.

%
% other terms
%
group_related_vars(Goal,Old_groups,Var_groups) :-
	functor(Goal,_,Arity),
	Arity>0,
	!,
	Goal =.. [_|Args],
	group_related_vars(Args,Old_groups,Var_groups).
*/

group_related_vars(_,Groups,Groups).



%
% add_var_into_group/3
%
% add a list of related vars into the list of groups
%

%
% no group so far
%
add_var_into_group(Vars,[],[Vars]).

add_var_into_group(Vars,[G|Gs],[New_g|Gs]) :-
	member(X,Vars),			%a var X in Vars
	var_member(X,G),		%is already in the group G
	!,
	var_append(Vars,G,New_g).	%append Vars into G
	
add_var_into_group(Vars,[G|Gs],[G|New_gs]) :-
	add_var_into_group(Vars,Gs,New_gs).

%
% append 2 lists of vars together
%
var_append([],Y,Y).

var_append([X|Xs],Y,Z) :-
	var_member(X,Y),
	!,
	var_append(Xs,Y,Z).

var_append([X|Xs],Y,[X|Temp]) :-
	var_append(Xs,Y,Temp).


%
% test for membership of var in a list
%
var_member(X,[H|_]) :-
	X == H,
	!.

var_member(X,[_|T]) :-
	var_member(X,T).


%
% delete an uninstantiated var from a list of vars
%
var_delete(_,[],[]).

var_delete(Var,[V|Vs],Vs) :-
	Var==V,				%Var is in the head of the list
	!.

var_delete(Var,[_|Vs],Deleted) :-
	var_delete(Var,Vs,Deleted).


post_links(_,[]).
post_links(Shared,[Daplex:Linking_var|Rest]) :-
	nonvar(Linking_var),
	!,
	find_prolog_var(Linking_var,Shared,Prolog_var1),	%find the FD var
	find_prolog_var(Daplex,Shared,Prolog_var2),
	Prolog_var1=Prolog_var2,
	post_links(Shared,Rest).

post_links(Shared,[Daplex:Linking_var|Rest]) :-
	var(Linking_var),
	!,
	Daplex=Linking_var,
	post_links(Shared,Rest).


%whatever(Vs_copy,Group,Related_vs_copy).

whatever([],_,[]).
whatever([V|Vs],Group,[V|Rest]) :-
	var_member(V,Group),
	!,
	whatever(Vs,Group,Rest).
whatever([_|Vs],Group,Rest) :-
	whatever(Vs,Group,Rest).

:- set_debug(forall,on).
:- set_debug(if_then_else,on).
%:- [domain].
