%:- ['util/for_eclipse'].

%
% the ECLiPSe solver module
%
% This is the environment where a compiled CLP will execute.
% Here are the utility predicates that a generated CLP program
% can call:
%
% find_prolog_var(Daplex,Var_pairs,Prolog):
%	find the corresponding FD vars for a list of Daplex vars
%
% declare_vars(Daplex,Prolog)
%	declare FD vars for Daplex vars, given their classes
%
% set_domain(Daplex,Class,Prolog)
%	a constraint restricting the domain of the FD var to its class
%	can also be used to define & declare a FD var from a Daplex var
%
% all_violate(V,G,P)
%	a universally quantified constraint
%
% exist(V,G,P)
%	an existentially quantified constraint
%
% label_vars(Daplex_prolog_list)
%	label all FD vars in Daplex:Prolog list
%

/*
:- module(eclipse_solver,[	tell_solver/1,		%tell ECLiPSe a fact/clause
				ask_solver/1		%call a predicate in the solver module
			]).
*/

:- use_module(library(fd)).		%use the FD library
:- use_module(library(propia)).		%use the Propia library
:- ['../../prolog/util/is_daplex_variable'].

:- dynamic(solve/1).			%the top-level predicate
:- dynamic(post_constraints/2).		%clauses posting constraints on vars
:- dynamic(post_links/1).

%:- [domain].				%*** vars domain info.

%
% the 2 main interfacing predicates
%

tell_solver(Something) :-
	assert(Something).

ask_solver(Predicate) :-
	call(Predicate).





%
% find_prolog_var/3
%
% given a (list of) Daplex var, find the corresponding Prolog var
%
% find_prolog_var(+Daplex,+Var_pairs,-Prolog)
%
% Daplex	=	Daplex var to find
% Var_pairs	=	list of Daplex:Prolog pairs
% Prolog	=	the Prolog var in ECLiPSe
%

find_prolog_var(Daplex,Var_pairs,Prolog) :-
	member(Daplex:Prolog,Var_pairs),
	!.

find_prolog_var([],_,[]).

find_prolog_var([D|Ds],Var_pairs,[P|Ps]) :-
	is_daplex_variable:is_daplex_variable(D),
	find_prolog_var(D,Var_pairs,P),
	find_prolog_var(Ds,Var_pairs,Ps).

%
% declare_vars/2
%
% given a list of Daplex:Type, declare finite domain vars
% and returns a list of Daplex:Prolog
%
% declare_vars(+Var_with_type,-Var_with_prolog)
%
%

declare_vars([],[]).

declare_vars(Daplex:Type,Daplex:Prolog) :-
	set_domain(Daplex,Type,Prolog).		%set its domain

declare_vars([V|Vs],[D|Ds]) :-
	declare_vars(V,D),
	declare_vars(Vs,Ds).




%
% there exist values in the variables generated by G
% such that P is satisfied
%
exist(_,G,P) :-
	call(G),
	call(P).


%
% all values in V generated by G are violating P
%
all_violate(V,G,P) :-
	not_exist(V,G,P).

%
% there exist no value in variables generated by G
% such that P is satisfied
%
not_exist(V,G,P) :-
	copy_term((V,G,P),(V_copy,G_copy,P_copy)),
	call(G_copy),
	call(P_copy),
	!,
	findall(V_copy,label(V_copy),Violating_v),
	write('*** violating:'),
	write(Violating_v),nl,
	call(G),
	remove_values(Violating_v,V),
	write('***removed:'),
	write(V),nl.

not_exist(_,G,_) :-
	call(G).

remove_values([],_).
remove_values([V|Vs],Vars) :-
	remove_values2(V,Vars),
	remove_values(Vs,Vars).

remove_values2([],_).
remove_values2([V|Vs],[X|Xs]) :-
	X #\= V,
	remove_values2(Vs,Xs).



label_vars([]).

label_vars([_:V|Rest]) :-
	indomain(V),
	label_vars(Rest).

label([]).
label([H|T]) :-
	indomain(H),
	label(T).


%
% set_domain/3
%
% set the domain of Daplex to the entity class Class
%
% set_domain(+Daplex,+Class,-Prolog)
%
% Daplex	=	a daplex var
% Class		=	an entity class
% Prolog	=	the ECLiPSe FD var
%
set_domain(_,integer,Prolog) :-
	Prolog :: 1..32767.

set_domain(_,string,_) :-
	!.

set_domain(_,Class,Prolog) :-
	object(Class,Prolog) infers most.


clean_up :-
	retractall(solve(_)),
	retractall(declare_vars(_)),
	retractall(post_constraints(_)),
	retractall(post_constraint(_,_)),
	retractall(post_links(_)),
	!.

clean_up.

post_links(_,[]).

post_links(Shared,[Daplex:Linking_var|Rest]) :-
	nonvar(Linking_var),
	!,
	find_prolog_var(Linking_var,Shared,Prolog_var1),	%find the FD var
	find_prolog_var(Daplex,Shared,Prolog_var2),
	Prolog_var1=Prolog_var2,
	post_links(Shared,Rest).

post_links(Shared,[Daplex:Linking_var|Rest]) :-
	var(Linking_var),
	!,
	Daplex=Linking_var,
	post_links(Shared,Rest).

