%
% module filtering empty-slots from a CSP expression
%
:- module(filter_empty_slot,[	filter_empty_slot/3,
				filter_empty_slot/4
	]).

:- use_module('../../prolog/util/prolog_detect').
:- prolog_detect:prolog(Prolog),
	(	(	Prolog=sicstus,
			use_module(library(lists))
		);
		Prolog=swi;
		Prolog=eclipse;
		(	write('*** unsupported Prolog system***'),
			nl,
			fail
		)
	).
:- use_module('../../prolog/util/cif_util').


%
% filter_empty_slot/3
%
% remove empty-slots from a CIF expression
% (including group/2)
%
% filter_empty_slot(+Cif,+Empty_slots,-Filtered)
%
% Cif		=	the CIF for filtering
% Empty_slots	=	a list of empty-slot templates
% Filtered	=	the filtered CIF
%

filter_empty_slot(Cif,Empty_slots,Filtered) :-
	filter_can_unify(Cif,Empty_slots,Filtered,_).

filter_empty_slot(Cif,Empty_slots,Filtered,Vars_removed) :-
	filter_can_unify(Cif,Empty_slots,Filtered,Vars_removed).

%
% filter_can_unify/4
%
% filter any term which can unify with some templates
%
% filter_can_unify(+Orig,+Templates,-Filtered,-Vars_removed)
%
% Orig		=	the original expression
% Templates	=	a list of template (w/ uninstantiated vars)
% Filtered	=	the filtered expression
% Vars_removed	=	list of vars removed because of filtering
%

%
% nothing to filter
%
filter_can_unify(Cif,[],Cif,[]) :-
	!.

%
% an empty-list
%
filter_can_unify([],_,[],[]) :-
	!.

%
% a group/2
%
filter_can_unify(Cif,Templates,Filtered,Vars_to_be_removed) :-
	Cif=group(Vars,In_group),
	!,
	filter_can_unify(In_group,Templates,In_group_filtered,Vars_filtered),
%	Filtered_temp=group(Vars,In_group_filtered),
	my_subtract(Vars,Vars_filtered,Remaining_vars),
	my_subtract(Vars_filtered,Vars,Vars_to_be_removed),
	Filtered=group(Remaining_vars,In_group_filtered).


%
% if head of list can unify w/ a template
% remove it from the result
%
filter_can_unify([C|Cs],Templates,Filtered,Vars_to_be_removed) :-
	member(E,Templates),			%there is an empty-slot
	can_unify(C,E),				%which can unify with C
	!,
	find_vars_to_remove(C,Vars_filtered),	%find the list of vars removed
	filter_can_unify(Cs,Templates,Filtered,More_vars_filtered),
	my_union(Vars_filtered,More_vars_filtered,Vars_to_be_removed).

%
% otherwise, recursively work on each element
% (NOTE: if C cannot unify w/ any template, it will be returned as F)
%
filter_can_unify([C|Cs],Templates,[F|Fs],Vars_to_be_removed) :-
	!,
	filter_can_unify(C,Templates,F,Vars_filtered1),
	filter_can_unify(Cs,Templates,Fs,Vars_filtered2),
	my_union(Vars_filtered1,Vars_filtered2,Vars_to_be_removed).


filter_can_unify(Cif,Templates,Filtered,Vars_to_be_removed) :-
	Cif=restrict_subquery(all(Vars),Gen,Pred),
	!,
	filter_can_unify(Gen,Templates,Gen_filtered,Vars_filtered1),
	filter_can_unify(Pred,Templates,Pred_filtered,Vars_filtered2),
	my_union(Vars_filtered1,Vars_filtered2,All_vars_filtered),
	my_subtract(Vars,All_vars_filtered,Remaining_vars),
	my_subtract(All_vars_filtered,Vars,Vars_to_be_removed),
	Filtered=restrict_subquery(all(Remaining_vars),Gen_filtered,Pred_filtered).

filter_can_unify(Cif,Templates,Filtered,Vars_to_be_removed) :-
	Cif=restrict_subquery(some(Vars),Gen,Pred),
	!,
	filter_can_unify(Gen,Templates,Gen_filtered,Vars_filtered1),
	filter_can_unify(Pred,Templates,Pred_filtered,Vars_filtered2),
	my_union(Vars_filtered1,Vars_filtered2,All_vars_filtered),
	my_subtract(Vars,All_vars_filtered,Remaining_vars),
	my_subtract(All_vars_filtered,Vars,Vars_to_be_removed),
	Filtered=restrict_subquery(some(Remaining_vars),Gen_filtered,Pred_filtered).


%
% any other term
%
filter_can_unify(Cif,Templates,Filtered,Vars_to_be_removed) :-
	functor(Cif,_,Arity),
	Arity>0,
	Cif =.. [Functor|Args],
	!,
	filter_can_unify(Functor,Templates,Filtered_functor,Vars_filtered1),
	filter_can_unify(Args,Templates,Filtered_args,Vars_filtered2),
	Filtered =.. [Filtered_functor|Filtered_args],
	my_union(Vars_filtered1,Vars_filtered2,Vars_to_be_removed).

%
% C cannot unify with any template
%
filter_can_unify(C,_,C,[]) :-
	!.


%
% can_unify/2
%
% test unifiability of 2 terms but not unifying them
%
can_unify(A,B) :-
	copy_term(A,A_clone),   %make a copy of A
	copy_term(B,B_clone),   %make a copy of A
	A_clone=B_clone.        %unifying the copy with B

%
% find the list of vars to remove when a CIF construct is filtered out
%
find_vars_to_remove(generate(_,Var),[Var]) :-
	!.
find_vars_to_remove(_,[]).
