%
% AKT-Bus <-> KRAFT gateway
%
% takes command line arguments:
%
% -my_url	URL to which the incoming gateway listens for AKT messagse
% -lh		Linda server host
% -lp		Linda server port
% -mode		Operation mode, a2k or k2a.
%

:- ['../../prolog/util/ksl'].				%load KRAFT service layer
:- ['../../prolog/util/cmdline'].			%load command line module
:- ['../../prolog/messaging/messaging'].		%load KRAFT messaging layer
:- ['../../prolog/util/prolog_detect'].
:- ['../../prolog/util/flatten'].
:- ['../../prolog/util/cif2xml'].			%load the CIF<->XML module
:- ['../../prolog/util/print_xml'].			%load the CIF<->XML module
:- ['~/quail_local/aktbus/prolog/src/absl'].		%load AKT-Bus Prolog library
:- prolog_detect:prolog(Prolog),
	(	(	Prolog=sicstus,
			use_module(library(lists))
		);
		(	Prolog=pfdm,
			use_module(library(lists))
		);
		true
	).

:- dynamic(to_close/1).


%
% URL of the PC config schema
%
url(domain_rdfs,'http://www.csd.abdn.ac.uk/~khui/akt/ix/pcconfig').

%
% TOP LEVEL PREDICATE
%
go :-
	ksl:read_agent_config,				%read config
	cmdline:get_command_line('-mode',Mode),		%read mode
	member(Mode,[a2k,k2a]),				%2 valid modes
	!,
	go(Mode).

%
% print usage
%
go :-
	write('*** go(a2k) runs AKT->KRAFT gateway'),
	nl,
	write('*** go(k2a) runs KRAFT->AKT gateway'),
	nl,
	write('*** or use the command line argumnet -mode'),
	nl.

go(Mode) :-
	ksl:get_config(linda_host,Host),
	ksl:get_config(linda_port,Port),
	start(Mode,Host,Port).

start(Mode,Host,Port) :-
	init(Mode,Host,Port,Aktbus_receiver),	%init
	main_loop(Mode,Aktbus_receiver),	%loop
	shutdown(Mode,Aktbus_receiver).		%shutdown

%
% initialise gateway
%

%
% AKT->KRAFT
%
init(a2k,Host,Port,Aktbus_receiver) :-
	write('... initialising AKT->KRAFT gateway'),
	write('... connecting to Linda server@ '),
	write(Host),
	write(':'),
	write(Port),
	nl,
	messaging:init_messaging([server_host:Host,server_port:Port]),
	write('... creating AKT-Bus receiver@ '),
	cmdline:get_command_line('-my_url',My_url),		%get URL that I'm listening to
	write(My_url),
	nl,
	absl:create_aktbus_receiver(My_url,Aktbus_receiver),	%create AKT-Bus receiver
	asserta(to_close(Aktbus_receiver)),
	!.

%
% KRAFT->AKT
%
init(k2a,Host,Port,_) :-
	write('... initialising KRAFT->AKT gateway'),
	nl,
	write('... connecting to Linda server@ '),
	write(Host),
	write(':'),
	write(Port),
	nl,
	messaging:init_messaging([server_host:Host,server_port:Port]),
	!.

init(_,_,_,_) :-
	write('*** FAIL to initialise gateway!!!'),
	nl,
	fail.



%
% shutting down
%
shutdown(a2k,Aktbus_receiver) :-
	write('... shutting down AKT->KRAFT gateway'),
	nl,
	write('... disconnecting from Linda server'),
	nl,
	messaging:halt_messaging,
	write('... destroying AKT-Bus receiver'),
	nl,
	absl:destroy_aktbus_receiver(Aktbus_receiver),
	!.

shutdown(k2a,_) :-
	write('... shutting down KRAFT->AKT gateway'),
	nl,
	write('... disconnecting from Linda server'),
	nl,
	messaging:halt_messaging,
	!.

%
% main loop
%
main_loop(a2k,Aktbus_receiver) :-
	repeat,								%loop from here
	write('... waiting for incoming AKT-Bus message'),
	nl,
	absl:receive_message(Aktbus_receiver,Akt_message),		%receive incoming AKT-Bus message
	write('... mapping AKT message to KRAFT message'),
	nl,
	write('... AKT message'),
	nl,
	print_xml:print_xml(Akt_message),
	nl,
	map_message(a2k,Akt_message,Kraft_message),			%map AKT message to KRAFT messge
	write('... sending KRAFT message'),
	nl,
	send_to_kraft(Kraft_message),					%send to KRAFT
	fail.								%fail to loop

main_loop(k2a,_) :-
	repeat,								%loop from here
	write('... waiting for incoming KRAFT message'),
	nl,
	messaging:receive_message(foreign(_),Kraft_message),		%receive message to foreign/1
	write('... mapping KRAFT message to AKT message'),
	nl,
	map_message(k2a,Akt_message,Kraft_message),			%map AKT message to KRAFT messge
	print_xml:print_xml(Akt_message),
	nl,
	write('... sending AKT message'),
	nl,
	send_to_akt(Akt_message),					%send to AKT
	fail.								%fail to loop

%
% send message to KRAFT
%
send_to_kraft(Kraft_message) :-
	messaging:send_message(Kraft_message),				%send to KRAFT
	!.

%
% send message to AKT
%
send_to_akt(Akt_message) :-
	absl:send_message(Akt_message),			%send OK
	!.

send_to_akt(Akt_message) :-				%sending fails
	write('*** FAIL sending message'),
	nl.



%
% map AKT-Bus message to KRAFT message
%
map_message(a2k,A_message,K_message) :-
	absl:decompose_message(A_sender,A_receiver,A_date,A_time,A_msgid,A_content,_,A_message),	%decompose AKT message
	%
	% map sender & receiver
	%
	map_sender(a2k,A_sender,K_sender),
	map_receiver(a2k,A_receiver,K_receiver),
	%
	% map date time
	%
	map_date(a2k,A_date,K_date),
	map_time(a2k,A_time,K_time),
%	K_date=date(Year,Month,Day),
%	K_time=time(Hour,Min,Sec),
	K_timestamp=time_stamp(K_date,K_time),
	%
	% map content
	%
	map_content(a2k,A_content,K_core),							%map AKT content into KRAFT core
	messaging:mf_compose_context(_,A_msgid,K_sender,K_receiver,K_timestamp,K_context),	%compose KRAFT message context
	messaging:mf_compose_message(K_context,K_core,K_message).				%compose KRAFT message

%
% map a KRAFT message to an AKT message
%
map_message(k2a,A_message,K_message) :-
	messaging:decompose_message(K_context,K_core,K_message),				%decompose KRAFT message
	messaging:decompose_context(_,K_id,K_sender,K_receiver,K_timestamp,K_context),		%decompose KRAFT context
	K_timestamp=time_stamp(K_date,K_time),
%	absl:decompose_message(A_sender,A_receiver,A_date,A_time,A_msgid,A_content,A_extra,Am),	%decompose AKT message
%	decompose_ccql_core(),
	map_sender(k2a,A_sender,K_sender),
	map_receiver(k2a,A_receiver,K_receiver),
	map_date(k2a,A_date,K_date),
	map_time(k2a,A_time,K_time),
	map_content(k2a,A_content,K_core),
	Name_spaces=[	cif_rdfs='http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-rdfs.xml',
			cif_compact='http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-compact-rdfs.xml',
			cif_classic='http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-classic-rdfs.xml'
			],
	absl:compose_message(A_sender,A_receiver,A_date,A_time,A_content,Name_spaces,A_message).		%compose AKT message

%
% map sender
%

%
% convert an AKT sender URL into a KRAFT foreign agent name
%
map_sender(a2k,A_sender,K_sender) :-
	name(A_sender,A_sender_string),					%convert URL from atom to string
	absl:my_http_url(url(Host,Port,Document),A_sender_string,[]),	%parse URL
	K_sender=foreign(url(Host,Port,Document)).			%make up a term

%
% convert a KRAFT local agent name into an AKT sender URL
%
map_sender(k2a,A_sender,K_sender) :-
	name(K_sender,K_sender_string),					%convert KRAFT agent name into string
	cmdline:get_command_line('-my_url',My_url),			%get URL of incoming gateway
	name(My_url,My_url_string),					%convert URL to string
	flatten:flatten([My_url_string,"/",K_sender_string],A_sender_string),	%concat strings together
	name(A_sender,A_sender_string).					%convert string to AKT-Bus URL

%
% map receiver
%

%
% convert an AKT receiver URL into a KRAFT foreign agent name
%
map_receiver(a2k,A_receiver,K_receiver) :-
	cmdline:get_command_line('-my_url',My_url),			%get incoming gate's URL
	name(My_url,My_url_string),					%convert atom to string
	absl:my_http_url(url(Host,Port,My_url_doc),My_url_string,[]),	%parse it

	name(A_receiver,A_receiver_string),					%convert AKT receiver into string
	absl:my_http_url(url(Host,Port,A_receiver_doc),A_receiver_string,[]),	%parse it
	name(My_url_doc,My_url_doc_string),
	name(A_receiver_doc,A_receiver_doc_string),
	append(My_url_doc_string,Agent_string,A_receiver_doc_string),		%get agent name in a string
	(	(	Agent_string=[0'/|Agent_string_no_slash],		%leading slash
			K_receiver_string=Agent_string_no_slash
		);
		K_receiver_string=Agent_string					%no leading slash
	),
	!,
	name(K_receiver,K_receiver_string).



%
% convert a KRAFT foreign receiver into a AKT receiver URL
%
map_receiver(k2a,A_receiver,K_receiver) :-
	K_receiver=foreign(url(Host,Port,Document)),			%KRAFT receiver is a foreign term
	name(Host,Host_string),
	(	(	Port=80,					%default port
			Port_string=""					%no need to specify
		);
		(	name(Port,Temp_port_string),
			append(":",Temp_port_string,Port_string)	%add ":" before port no.
		)
	),
	name(Document,Temp_document_string),
	(	(	Temp_document_string=[0'/|_],			%start w/ a "/"
			Document_string=Temp_document_string		%ok
		);
		Document_string=[0'/|Temp_document_string]		%otherwise, add "/" to the beginning
	),
	!,
	flatten:flatten(["http://",Host_string,Port_string,Document_string],A_receiver_string),	%concat all strings together
	name(A_receiver,A_receiver_string).				%convert AKT receiver URL string to atom

map_date(_,Date,Date).

map_time(_,Time,Time).

%
% map AKT message content to KRAFT
%
map_content(a2k,A_content,K_core) :-
	A_content=env('akt-kqml:KqmlContent',_,A_params),				%get akt-kqml:Content & KQML params
	map_params(a2k,A_params,Param_list),						%get param:value list
	member('hasPerformative':Perf,Param_list),					%look for performative
	delete(Param_list,'hasPerformative':Perf,Param_list2),				%remove performative from param list
	messaging:mf_set_argument(Param_list2,K_params_orig),				%make KRAFT style params
	messaging:get_argument(K_params_orig,language:Language),			%get language of content
	messaging:get_argument(K_params_orig,content:Rdf_constraint),			%get content, which maybe a constraint
	%
	% map constraints from XML -> Prolog
	%
%	Prolog_constraint=[xmldecl([version='1.0']),Rdf_constraint],		%put XML header back for constraint parsing
	Xml_constraint=[xmldecl([version='1.0']),env('rdf:RDF',[],[Rdf_constraint])],		%put XML header back for constraint parsing
	map_constraint(Language,xml,prolog,Xml_constraint,Prolog_constraint),	%map constraint XML->Prolog if needed
	%
	% empty slots
	%
	Empty=[generate(allocate,_),
		generate(arg1,[allocate],_,_,_),
		generate(arg2,[allocate],_,_,_)
		],
	cif_util:find_var_types(Prolog_constraint,Types),		%find all vars in var:type form
	findall(V,member(V:_,Types),Output_vars),
	findall(T,member(_:T,Types),Output_types),
	%
	% content pattern that the KRAFT FM understands
	%
	Pattern=[	Output_vars:Output_types,
			generate(pc_config,[set_of(cif),set_of(cif)],[Prolog_constraint,Empty],Output_types,Output_vars)
		],
	messaging:set_argument(K_params_orig,content:Pattern,K_params),		%make up pattern
	messaging:compose_ccql_core(Perf,K_params,K_core).				%compose KRAFT message core

%
% map KRAFT message content to AKT
%
map_content(k2a,A_content,K_core) :-
	messaging:decompose_ccql_core(Perf,K_params,K_core),			%decompose CCQL core
	findall(X:Y,messaging:get_argument(K_params,X:Y),Colon_param_list),	%get list of param in X:Y format
%	member(language:Language,Colon_param_list),
	member(content:Something,Colon_param_list),				%find content
	delete(Colon_param_list,content:Something,Temp_colon_params),
	%
	% map constraint into XML
	%
	write('*** content to map:'),
	nl,
	write(Something),
	nl,
	write('***'),
	nl,
%	map_constraint(Language,prolog,xml,Something,Xml_something),		%map something from Prolog -> XML
	(	(Something=solve(Vars,Tuple_set),Tuple_set=[[_|_]|_],To_map=[Vars:_,Tuple_set]);
		(Something=solve(Vars,Tuple),To_map=[Vars:_,[Tuple]]);
		To_map=Something
	),
	!,
	map_result(Language,prolog,xml,To_map,Xml_something),
	%
	% take body only, without XML header
	%
%	Rdf_something=env('rdf:RDF',_,_),					%matching template
%	member(Rdf_something,Xml_something),
%	New_k_params=[content:Rdf_something|Temp_colon_params],
	New_k_params=[content:Xml_something|Temp_colon_params],
	%
	% map colon param list into AKT style
	%
	map_params(k2a,Temp_a_params,New_k_params),				%map colon params into AKT params
	Perf_param=env('akt-kqml:hasPerformative',[],[Perf]),			%the performative in AKT KQML
	A_params=[Perf_param|Temp_a_params],					%add perf into AKT param list
	%
	% make up XML KQML
	%
	A_content=env('akt-kqml:KqmlContent',[],A_params).			%the content

map_params(a2k,[],[]).
map_params(a2k,[A|As],[K|Ks]) :-
	A=env(Tagname,[],[Param_value]),
	name(Tagname,Tagname_string),					%convert RDF tag name to string
	append("akt-kqml:",Param_name_string,Tagname_string),		%take away the prefix "akt-kqml:"
	name(Param_name,Param_name_string),
	K=Param_name:Param_value,
	map_params(a2k,As,Ks).

map_params(k2a,[],[]).
map_params(k2a,[A|As],[K|Ks]) :-
	K=Param_name:Param_value,
	name(Param_name,Param_name_string),
	append("akt-kqml:",Param_name_string,Tagname_string),
	name(Tagname,Tagname_string),
	A=env(Tagname,[],[Param_value]),
	map_params(k2a,As,Ks).

%
% map constraint from XML<->Prolog format
%



%
% no need to map
%
map_constraint(cif,prolog,prolog,Constraint,Constraint) :-
	!.
map_constraint(cif,xml,xml,Constraint,Constraint) :-
	!.
map_constraint(cif,xml,prolog,Xml_constraint,Prolog_constraint) :-
	url(domain_rdfs,Domain_rdfs),
	Params=[	representation=classic,
			domain_rdfs=Domain_rdfs,
			constraint_id=generate],
	(	cif2xml:cif2xml(Params,Prolog_constraint,Xml_constraint);
		cif2xml:query2xml(Params,Prolog_constraint,Xml_constraint);
		cif2xml:result2xml(Params,Prolog_constraint,Xml_constraint)
	),
	!.


map_constraint(cif,prolog,xml,Prolog_constraint,Xml_constraint) :-
	url(domain_rdfs,Domain_rdfs),
	Params=[	representation=classic,
			domain_rdfs=Domain_rdfs,
			constraint_id=generate],
	(	cif2xml:cif2xml(Params,Prolog_constraint,Xml_constraint);
		cif2xml:query2xml(Params,Prolog_constraint,Xml_constraint);
		cif2xml:result2xml(Params,Prolog_constraint,Xml_constraint)
	),
	!.

%
% not a CIF, dont' know how to map
%
map_constraint(_,_,_,Constraint,Constraint) :-
	!.

%
% map result tuples from Prolog to XML
%
map_result(Language,prolog,xml,Something,Xml_result) :-
%	Tuples=result(tuples(Something)),
	url(domain_rdfs,Domain_rdfs),
	Params=[	representation=classic,
			domain_rdfs=Domain_rdfs,
			constraint_id=generate],
	(	Something=[Template,Set];
		Something=[[Template,Set]]		%*** I don't know why there is an extra bracket here!
	),
	Template=Vars:_,
	map_tuples(Vars,Set,Tuples),
	cif2xml:result2xml(Params,result(tuples(Tuples)),Xml_something),
	Xml_something=[xmldecl([version='1.0']),env('rdf:RDF',_,Xml_result)].	%remove XML header

map_tuples(_,[],[]).
map_tuples(Vars,[S|Ss],[T|Ts]) :-
	map_tuple(Vars,S,T),
	map_tuples(Vars,Ss,Ts).

map_tuple([],[],[]).
map_tuple([V|Vs],[X|Xs],[V=X|Rest]) :-
	map_tuple(Vs,Xs,Rest).

quit :-
	to_close(Aktbus_receiver),
	absl:destroy_aktbus_receiver(Aktbus_receiver).

