%
% KRAFT-IX adaptor
%
% This "mediator" maps IX knowledge to KRAFT format.
%
% takes command line parameters:
%
% -mode		operation mode, k2i/i2k
% -my_url	URL of this KIA agent
% -kg_url	URL of the KRAFT-AKT gateway
% -ax		Linda address XML file URL
%

:- ['../../prolog/util/cmdline'].			%load the command line module
:- ['../../prolog/util/print_xml'].			%load the print XML module
:- ['../../prolog/util/ksl'].
:- ['../../prolog/messaging/messaging'].		%load the KRAFT messaging layer
:- ['~/quail_local/aktbus/prolog/src/absl'].		%load the AKT-Bus service layer
%:- ['../../prolog/util/cif2xml'].
:- use_module(library(charsio)).
:- use_module(library(lists)).
:- use_module(library(system)).
:- [ix2kraft].						%the IX->KRAFT mapping module

:- dynamic(url/2).
:- dynamic(to_close/1).

go :-
	init(Receiver),
	loop(Receiver),
	stop(Receiver).

%
% init agent
%
init(Receiver) :-
	ksl:read_agent_config,
	write('... initialising IX->KRAFT gateway'),
	nl,
	cmdline:get_command_line('-my_url',My_url),		%URL where I receive incoming messages
%	cmdline:get_command_line('-kg_url',Kraft_url),		%URL of KRAFT gateway entry point
	assertz(url(my_url,My_url)),
%	assertz(url(kg_url,Kraft_url)),
	write('...connecting to KRAFT Linda'),
	nl,
	ksl:get_config(linda_host,Host),
	ksl:get_config(linda_port,Port),
	messaging:init_messaging([server_host:Host,server_port:Port]),
	write('... creating AKT-Bus receiver@'),
	write(My_url),
	nl,
	create_aktbus_receiver(My_url,Receiver),
	asserta(to_close(Receiver)),
	!.

init(_) :-
	write('*** FAIL to init KRAFT<->IX adaptor'),
	nl,
	fail.

%
% main loop
%
loop(Receiver) :-
	url(my_url,My_url),						%get my URL
	repeat,							%loop from here
	write('... waiting for AKT message@'),
	write(My_url),
	nl,
	absl:receive_message(Receiver,Message),
	handle_message(Message),
	fail.							%fail to loop


%
% stop agent
%
stop(Receiver) :-
	absl:destroy_aktbus_receiver(Receiver).

%========================
% message handler section
%========================

%
% known URLs
%
known_url(rdf,'http://www.w3.org/1999/02/22-rdf-syntax-ns').
known_url(rdfs,'http://www.w3.org/2000/01/rdf-schema').
known_url(aktmessage_rdfs,'http://www.csd.abdn.ac.uk/~khui/akt/aktbus.xml').
known_url(issue_rdfs,'http://www.csd.abdn.ac.uk/~khui/akt/ix/issue-rdf.xml').

%
% given a name (I know) OR a URL, find its namespace "name" in the current RDF
% We need to do this because namespace "name" can change but URL will not (assumed).
%
get_namespace(Name,Url,Namespace) :-
	known_url(Name,Url),
	concat_atoms(Url,'#',Url_with_hash),
	rdf2triple:namespace(Namespace,Url_with_hash).

%
% handle received messages
% succed to loop
%
handle_message(Message) :-
	write('*** message received:'),nl,
	print_xml:print_xml(Message),
	nl,
	%
	% parse RDF message as triples
	%
	rdf2triple:clean_triple,		%clean old triples
	rdf2triple:rdf2triple(Message),		%parse RDF message into triples
	%
	% now get issue_packet in message
	%
	get_namespace(rdf,Rdf_namespace),
	get_namespace(issue_rdfs,Issue_namespace),
	rdf2triple:triple(Issue_packet_resource,Rdf_namespace:'RDF',Issue_namespace:issue_packet),	%find issue_packet resource
%	[Issue_namespace:issue_packet_issue_instance,
%	Issue_namespace:issue_content,
%	]

	get_value(Issue_packet_resource,[],Value),
	true.

%
% follow a list of predicates to find a value
%
get_value(Root,[P],Value) :-
	get_predicate_value(direct,Root,P,Value),
get_value(Root,[P|Ps],Value) :-
	get_predicate_value(direct,Root,P,Intermediate),
	get_value(Intermediate,Ps,Value).


%
% find the value of a predicate under a certain node
%
get_predicate_value(Direct_or_indirect,Root,Predicate,Value) :-
	get_predicate_value([],Root,Predicate,Value) :-

get_predicate_value(_,History,Root,Predicate,Value) :-
	rdf2triple:triple(Root,Predicate,Value).

get_predicate_value(indirect,History,Root,Predicate,Value) :-
	rdf2triple:triple(Root,_,Intermediate),		%get an intermediate resource
	not(member(Intermediate,History)),		%which has not seen before
	get_predicate_value([Intermediate|History],Intermediate,Predicate,Value).	%1 more node in history



%
% if content is encoding as XML
%
handle_message2(Message) :-
	absl:get_env(Message,'issue',_,[Issue_subtree]),
	write('...got <issue> value'),
	absl:get_env(Issue_subtree,'statement',_,[Statement_atom]),
	write('...got <statement> value'),
	nl,
	%
	% convert atom back to term
	%
	name(Statement_atom,Statement_string),					%convert atom to a strig
	append(Statement_string,".",Statement_string_with_fullstop),		%need a fullstop at the end
	read_from_chars(Statement_string_with_fullstop,Statement),		%read term from string
	write('...got <statement> as a term'),
	nl,
	Statement=post_issue((_Sender,_Receiver),Issue),
	Issue=(_,_,_,_,_,Solutions),
	member(solution(_,_,Old_solution,_,_),Solutions),
	(Constraints=Old_solution;
	member(Constraints,Old_solution)
	),
	write('...got constraints'),
	nl,
	write(Constraints),
	nl,
%	extract_content(Message,Content),
%	map_ix_to_kraft(Content,Kraft_content),
%	send_to_kraft(Kraft_content),
	!,
	fail.

%
% SP's encoding where the whole thing is a string/atom
%
handle_message2(Message) :-
	absl:get_sender_uri(Message,Sender),
	absl:get_receiver_uri(Message,Receiver),
/*
	write('...getting aktbus:SimpleContent...'),
	nl,
	absl:get_env(Message,'aktbus:SimpleContent',_,[Simple_content_subtree]),
	write('...getting aktbus:hasSimpleMessage...'),
	nl,
	absl:get_env(Simple_content_subtree,'aktbus:hasSimpleMessage',_,[Simple_content_atom]),
	name(Simple_content_atom,Simple_content_string),
	write('...removing XML entity references from string...'),
	nl,
	map_xml_string(String_to_parse,Simple_content_string,[]),
	write('...parsing string into XML...'),
	nl,
	absl:string_to_xml(String_to_parse,Simple_content_xml_tree),
	print_xml:print_xml(Simple_content_xml_tree),
	nl,
	write('...looking for <issue> in XML tree...'),
	nl,
	absl:get_env(Simple_content_xml_tree,'issue',_,Issue_tree),
	write('...looking for <string> in <issue> subtree...'),
	nl,
	absl:get_env(Issue_tree,'string',_,[Issue_packet_atom]),
	write('...converting issue_packet atom to string...'),
	nl,
	name(Issue_packet_atom,Issue_packet_string),
	write('...converting issue_packet string to term...'),
	nl,
	make_sure_full_stop_at_end(Issue_packet_string,Sure_with_full_stop),
	read_from_chars(Sure_with_full_stop,Issue_packet_term),
	write('...got issue_packet/2'),
	nl,
	Issue_packet_term=issue_packet(_,Issue_occ_term),
	Issue_occ_term=issue_occ(_,_,_,_,_,_,Solution_term),
	write('...got issue_occ/7'),
	nl,
	write('...got solution/5'),
	nl,
	Solution_term=solution(_,_,That_damn_thing,_,_),
	(	Ix_constraints=That_damn_thing;
		member(Ix_constraints,That_damn_thing)
	),
	write(Ix_constraints),
	nl,
%	write('...mapping IX constraint to KRAFT...'),
%	nl,
%	ix2kraft:map_constraint(Ix_constraints,Kraft_constraints),
%	write('***KRAFT constraint:'),
%	nl,
%	print_cif:print_cif(Kraft_constraints),
%	nl,
%	make_up_kqml_ask(Kraft_constraints,Kqml_message),
*/
	make_up_kqml_ask(Kqml_message),
	messaging:send_message(Kqml_message),
%	sleep(1),
	url(my_url,My_url),
	name(My_url,My_url_string),					%convert it to string
	absl:my_http_url(My_url_term,My_url_string,[]),			%parse URL
	messaging:receive_message(foreign(My_url_term),Reply),
	write('*** reply from KRAFT'),
	nl,
	write(Reply),
	nl,
	write('...sending to IX'),
	nl,
	extract_kraft_reply(Reply,Reply_content),
	send_to_ix(Receiver,Sender,Reply_content),
	!,
	fail.
/*
send_to_kraft(Kraft_constraints) :-
	url(my_url,My_url),
	url(kg_url,Kraft_url),
	messaging:compose_message(),			%compose KRAFT message
	messaging:send_message(Message).		%send to KRAFT
%	absl:compose_message(My_url,Kraft_url,Kraft_constraints,Message),
%	absl:send_message(Message).
*/


%
% remove entity reference (i.e. '&...') in a string
% so that it can be parse by Pillow's XML parser
%

map_xml_string([]) -->
	"".
map_xml_string([C|Cs]) -->
	map_xml_entity(C),
	map_xml_string(Cs).

map_xml_entity(0'>) -->
	"&gt;",{!}.
map_xml_entity(0'<) -->
	"&lt;",{!}.
map_xml_entity(0'&) -->
	"&amp;",{!}.
map_xml_entity(0'") -->
	"&quot;",{!}.
map_xml_entity(0' ) -->
	"&nbsp;",{!}.
map_xml_entity(C) -->
	[C].


make_sure_full_stop_at_end(String,String) :-
	append(_,".",String),
	!.

make_sure_full_stop_at_end(String,Sure) :-
	append(String,".",Sure).

quit :-
	to_close(Receiver),
	absl:destroy_aktbus_receiver(Receiver).


make_up_kqml_ask(Kraft_constraints,Content) :-
	Content=env('akt-kqml:KqmlContent',[],[Params]),
	Params=[Performative_param,
		Sender_param,
		Receiver_param,
		Language_param,
		Ontology_param,
		Reply_with_param,
		Content_param
		],
	Performative_param=env('akt-kqml:hasPerformative',[],[ask_one]),
	url(my_url,Me),
	url(kg_url,Kraft_gateway),
	Sender_param=env('sender',[],[Me]),
	Receiver_param=env('receiver',[],[Kraft_gateway]),
	Language_param=env('language',[],[xcif]),
	Ontology_param=env('ontology',[],[shared]),
	Reply_with_param=env('reply_with',[],[this]),
	write_to_chars(Kraft_constraints,String),
	name(Constraint_atom,String),
	Content_param=env('content',[],[Constraint_atom]),

	write('KQML content:'),
	nl,
	print_xml:print_xml(Content).

%
% makeup one
%
make_up_kqml_ask(Kqml_message) :-
	cmdline:get_command_line('-my_url',My_url),			%get my URL
	name(My_url,My_url_string),					%convert it to string
	absl:my_http_url(My_url_term,My_url_string,[]),			%parse URL
	Sender=foreign(My_url_term),					%I'm the sender
	Receiver=fm,							%fuser is the receiver
	Reply_with=test_123,

	findall(C,get_constraint(C),All_constraints),			%get all constraints
	%
	% compose message
	%
	Cif=All_constraints,
	%       cif_util:find_var_types(All_constraints,Vars_with_types),
	%       findall(V,member(V:_,Vars_with_types),Output_vars),
	%       findall(T,member(_:T,Vars_with_types),Output_types),

	Output_vars=[var(bx),var(sx)],
	Output_types=[board,slot],

	Empty=[ generate(allocate,_),
	generate(arg1,[allocate],_,_,_),
	generate(arg2,[allocate],_,_,_)
	],
	Pattern=[       Output_vars:Output_types,
	generate(pc_config,[set_of(cif),set_of(cif)],[Cif,Empty],Output_types,Output_vars)
	],
	Perf=ask_all,
	messaging:set_argument([        sender:Sender,
					receiver:Receiver,
					reply_with:Reply_with,
					content:Pattern,
					ontology:shared,
					language:xcif
				],Params),
	messaging:compose_message(Sender,Receiver,Perf,Params,Kqml_message).


%
% extract reply content from reply
%
extract_kraft_reply(Reply,Reply_content) :-
	messaging:decompose_message(_,_,_,Params,Reply),
	messaging:get_argument(Params,content:Reply_content).

send_to_ix(Sender,Receiver,Content) :-
	write('...sending reply to IX'),
	nl,
	write_to_chars(Content,String),
	name(Content_atom,String),
	absl:compose_message(Sender,Receiver,Content_atom,Message),
	absl:send_message(Message).

%
% CONSTRAINTS
%

/*
get_constraint(C) :-
	C=restrict_subquery(some([var(ax),var(bx),var(px),var(sx)]),
		[
			generate(allocate, var(ax)),
			generate(board, var(bx)),
			generate(processor, var(px)),
			generate(slot, var(sx)),
			generate(arg1, [allocate], [var(ax)], board, var(bx)),
			generate(arg2, [allocate], [var(ax)], slot, var(sx)),
			relative(board, var(px), var(bx))
		],
		[icode_true]
		).
*/

/*
get_constraint(C) :-
	C=restrict_subquery(some([var(12), var(11), var(10), var(9), var(8), var(7), var(6), var(5), var(4), var(3), var(2), var(1)]),
		[	generate(processor, var(1)),
			generate(disk_controller, var(2)),
			generate(io, var(3)),
			generate(allocate, var(4)),
			generate(arg1, [allocate], [var(4)], board, var(5)),
			relative(board, var(1), var(5)),
			generate(arg2, [allocate], [var(4)], slot, var(6)),
			generate(board, var(5)),
			generate(allocate, var(7)),
			generate(arg1, [allocate], [var(7)], board, var(8)),
			relative(board, var(2), var(8)),
			generate(arg2, [allocate], [var(7)], slot, var(9)),
			generate(board, var(8)),
			generate(allocate, var(10)),
			generate(arg1, [allocate], [var(10)], board, var(11)),
			relative(board, var(3), var(11)),
			generate(arg2, [allocate], [var(10)], slot, var(12)),
			generate(board, var(11)),
			generate(slot, var(6)),
			generate(slot, var(9)),
			generate(slot, var(12))
		],
		[icode_true]
		).
*/



get_constraint(C) :-
C=restrict_subquery(all([var(a),var(s),var(b),var(p),var(sl),var(pl)]),
	[	generate(allocate,var(a)),
		generate(processor,var(p)),
		generate(board,var(b)),
		relative(board,var(p),var(b)),
		generate(string,var(sl)),
		generate(string,var(pl)),
		generate(arg1,[allocate],[var(a)],board,var(b)),
		generate(arg2,[allocate],[var(a)],slot,var(s)),
		generate(length,[slot],[var(s)],string,var(sl)),
		generate(length,[processor],[var(p)],string,var(pl)),
		expression([],[],expr(=,var(pl),long))
	],
	[	expression([],[],expr(=,var(sl),long))
	]
	).


/*
get_constraint(C) :-
C=restrict_subquery(all([var(a2),var(s2),var(b2),var(p2),var(sp2),var(pp2)]),
	[	generate(allocate,var(a2)),
		generate(processor,var(p2)),
		generate(board,var(b2)),
		relative(board,var(p2),var(b2)),
		generate(string,var(sp2)),
		generate(string,var(pp2)),
		generate(arg1,[allocate],[var(a2)],board,var(b2)),
		generate(arg2,[allocate],[var(a2)],slot,var(s2)),
		generate(extra_power_data,[slot],[var(s2)],string,var(sp2)),
		generate(extra_power_data,[processor],[var(p2)],string,var(pp2)),
		expression([],[],expr(=,var(pp2),extra))
	],
	[	expression([],[],expr(=,var(sp2),extra))
	]
	).
*/


