%
% wrapper for P/FDM
%
% P/FDM has a different environment from SICStus as it is a compiled runtime.
% For example, it doesn't take SICStus' command line argument of loading a Prolog file.
% So we have to mimic the behaviour by ourselves.
%

:- use_module('../../prolog/util/ksl').
:- use_module('../../prolog/messaging/messaging').		%the messaging module
:- use_module('../../prolog/util/cmdline').			%the command-line module
:- use_module('../../prolog/util/is_daplex_variable').
:- use_module('../../prolog/util/cif_util').
:- use_module('../../prolog/util/cif2i').			%convert CIF to ICode
:- use_module(library(lists)).
:- use_module(library(system)).

%
% different starting predicates
%
% go		:	Read config file name from command-line arg "-c filename"
%			If no "-c" is given, use the default "wrapper.config".
%			Optional linda host & port given by "-lh" & "-lp".
%
% go(Host,Port)	:	Use the linda server at Host:Port.
%
go :-
	ksl:read_agent_config,
	ksl:get_config(linda_host,Host),
	ksl:get_config(linda_port,Port),
	start(Host,Port).

start(Host,Port) :-
	init(Host,Port),
	main_loop,
	shutdown.




%
% initializing the agent
%
init(Host,Port) :-
	format("*** initialising W...~n",[]),
	messaging:init_messaging([server_host:Host,server_port:Port]),
	local_init.

%
% shutting the agent
%
shutdown :-
	format("*** shutting down W...~n",[]),
	local_shutdown,
	messaging:halt_messaging.

%
% look for database_path & database_module
%
local_init :-
	get_config(database_path,Path),
	get_config(database_module,Module),
	set_directory(Module,Path),
	format("... P/FDM database directory set...~n",[]),
	open_module(Module,read),
	format("... P/FDM database module opened...~n",[]),
	!.

%
% otherwise look for database_command
%
local_init :-
	get_config(database_command,Goal),
	format("... Executing database initialization command...~n",[]),
	call(Goal),
	format("... Command executed...~n",[]),
	!.

local_shutdown :-
	get_config(database_module,Module),
	close_module(Module),
	!.

%
% the main loop
%
main_loop :-
	format("*** starting W main loop~n",[]),
	ksl:get_config(name,Me),
	repeat,
	write('... waiting for incoming message'),
	nl,
	receive_incoming(Me,Message),
	format("... something received...~n",[]),
	messaging:decompose_message(Sender,Me,Perf,Param,Message),
	format("... handling income message...~n",[]),
	handle_message(Sender,Me,Perf,Param,Message).		%MUST fail for looping

receive_incoming(Me,Message) :-
	messaging:receive_message(Me,Message),
	!.
%
% handling different messages
%

%
%PING
%
handle_message(Sender,Me,ping,Params,Message) :-
	format(">>> PING received~n",[]),
	messaging:decompose_message(Context,_,Message),
	messaging:decompose_context(Version,Cid,_,_,_,Context),
	Check_list=[if_not(reply_with:_,'reply_with not found')],
	check_params(Params,Check_list,_,Error),
	(	(	%
			% no problem
			%
			Error=none,
			!,
			messaging:get_argument(Params,reply_with:Reply_with),		%get reply_with
			messaging:set_argument(in_reply_to:Reply_with,Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(pong,Reply_params,Reply_core),	%compose PONG core
			format("... PONG composed~n",[])
		);
		(	messaging:set_argument(content:Error,Reply_params),
			messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose ERROR core
			format("... ERROR composed~n",[])
		)

	),
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	format("<<< sending reply~n",[]),
	send_message(Reply),
%	sleep(1),
	!,
	fail.


/*** FORGET ABOUT IT FOR THE MOMENT
%
% ASK_IF
%
handle_message(Sender,Me,ask_if,Params,Message) :-
	format(">>> ASK_IF received~n",[]),
	Check_list=[	if_not(language:cif,'language must be cif'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:_,'content missing')
			],
	check_params(Params,Check_list,_,Error),
	(	(	%
			% no error
			% and I can handle this ask_if
			%
			Error=none,
			messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
							content:Pattern]),		%and pattern
			match_capability(Pattern),					%within my capability
			!,
			messaging:set_argument([	in_reply_to:Reply_with,
							content:Pattern
						],
							Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(tell,Reply_params,Reply_core),	%compose TELL core
			format("... TELL composed~n",[])

		);
		(	%
			% no error, but I can't handle this ask_if
			%
			Error=none,
			!,
			messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
							content:Pattern]),		%and pattern
			messaging:set_argument([	in_reply_to:Reply_with,
							content:Pattern
							],
							Reply_params),	%use as in_reply_to
			messaging:compose_ccql_core(sorry,Reply_params,Reply_core),	%compose TELL core
			format("... SORRY composed~n",[])
		);
		(	messaging:set_argument(content:Error,Reply_params),
			messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
			format("... ERROR composed~n",[])
		)
	),
	messaging:decompose_message(Context,_,Message),
	messaging:decompose_context(Version,Cid,_,_,_,Context),
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	format("<<< sending reply~n",[]),
	send_message(Reply),
%	sleep(1),
	!,
	fail.
*/

%
% ASK_ONE
%
handle_message(Sender,Me,ask_one,Params,Message) :-
	format(">>> ASK_ONE received~n",[]),
	Check_list=[	if_not(language:cif,'language must be cif'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:[_,_],'content not in the form [Variables,CIF]')
			],
	check_params(Params,Check_list,_,Error),				%check for any error
	handle_ask_one(Sender,Me,Params,Error,Reply_core),			%handle ASK_ONE
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get versio & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	format("<<< sending reply~n",[]),
	send_message(Reply),
	format("... reply sent~n",[]),
%	sleep(1),
	!,
	fail.




%
% ASK_ALL
%
handle_message(Sender,Me,ask_all,Params,Message) :-
	format(">>> ASK_ALL received~n",[]),
	Check_list=[	if_not(language:cif,'language must be cif'),
			if_not(ontology:shared,'ontology must be shared'),
			if_not(reply_with:_,'reply_with not found'),
			if_not(content:[_,_],'content not in the form [Variables,CIF]')
			],
	check_params(Params,Check_list,_,Error),				%check for any error
	handle_ask_all(Sender,Me,Params,Error,Reply_core),			%handle ASK_ONE
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get versio & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	format("<<< sending reply~n",[]),
	send_message(Reply),
	format("... reply sent~n",[]),
%	sleep(1),
	!,
	fail.


%
% HALT
%
handle_message(_,_,halt,_,_) :-
	format("W>>> agent terminating...~n",[]),
	!.

%
% DEBUG
%
handle_message(_,_,debug,_,_) :-
	format("DEBUGGING invoked by message...~n",[]),
	send_register,
	send_advertise,
	!,
	fail.

%
% ALL OTHERS
%
handle_message(Sender,Me,Perf,Params,Message) :-
	format("*** WARNING: unknown performative received~n",[]),
	format("... composing SORRY~n",[]),
	%
	% compose SORRY to sender
	%
	name(Perf,Perf_string),
	flatten(["cannot handle ",Perf_string],Error_string),			%compose error message string
	name(Error_message,Error_string),
	Check_list=[if_not(reply_with:_,'reply_with not found')],		%check for params
	check_params(Params,Check_list,_,Error),
	(	(	Error=none,						%if no problem w/ reply-with
			!,
			messaging:get_argument(Params,[reply_with:Reply_with]),	%get reply-with
			messaging:set_argument([in_reply_to:Reply_with,		%compose reply params w/ in-reply-to & content
						content:Error_message],
						Reply_params)
			);
		messaging:set_argument([content:Error_message],Reply_params)	%otherwise just put in the content
	),
	messaging:compose_ccql_core(sorry,Reply_params,Reply_core),		%compose SORRY core
	messaging:decompose_message(Context,_,Message),				%get original context
	messaging:decompose_context(Version,Cid,_,_,_,Context),			%get version & cid
	messaging:compose_context(Version,Cid,Me,Sender,_,Reply_context),	%compose the context
	messaging:compose_message(Reply_context,Reply_core,Reply),		%compose message
	send_message(Reply),
	format("<<< SORRY sent~n",[]),
	flush_output(user),
	!,
	fail.


%
% check if params satisfy the required value
% return error message if appropriate
%
check_params(_,[],_:_,none).

check_params(Param,[C|Cs],Problematic_pair,Error) :-
	C=if_not(Arg:Expected_value,This_error),
	(	(	var(Expected_value),					%expecting any value
			messaging:get_argument(Param,Arg:Expected_value),	%found it there
			!,
			check_params(Param,Cs,Problematic_pair,Error)		%proceed
		);
		(	nonvar(Expected_value),					%expecting a certain value
			messaging:get_argument(Param,Arg:Current_value),	%get current one
			nonvar(Current_value),					%which is not a var
			Current_value=Expected_value,				%and they can unify
			!,
			check_params(Param,Cs,Problematic_pair,Error)
		);
		(	nonvar(Expected_value),
			messaging:get_argument(Param,Arg:Current_value),
			var(Current_value),
			!,
			Error=This_error,
			Problematic_pair=Arg:Current_value
		);
		(	Error=This_error,
			Problematic_pair=Arg:Expected_value
		)
	).


%
% handling different situations of ASK_ONE
%
%
handle_ask_one(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, and I can handle it
	%
	messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
					content:[Vars:Types,Query]]),	%get vars & query CIF
	process_ccql(ask_one,Vars,Types,Query,Result),
	messaging:set_argument([sender:Me,				%reply back
				receiver:Sender,
				in_reply_to:Reply_with,
				content:Result
				],
				Reply_params),				%use as in_reply_to
	messaging:compose_ccql_core(tell,Reply_params,Reply_core).	%compose TELL core

handle_ask_one(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, but problem in processing ICode
	%
	messaging:get_argument(Params,reply_with:Reply_with),		%get reply_with
	messaging:set_argument([sender:Me,
				receiver:Sender,
				in_reply_to:Reply_with,
				content:'problem in processing query'
				],
				Reply_params),
	messaging:compose_ccql_core(error,Reply_params,Reply_core).

handle_ask_one(Sender,Me,Params,Error,Reply_core) :-
	%
	% has error
	%
	(	(	messaging:get_argument(Params,[reply_with:Reply_with]),
			!,
			messaging:set_argument([sender:Me,
						receiver:Sender,
						in_reply_with:Reply_with,
						language_cif,
						ontology:shared,
						content:Error
						],Reply_params)
		);
		messaging:set_argument([sender:Me,
					receiver:Sender,
					in_reply_with:Reply_with,
					language_cif,
					ontology:shared,
					content:Error
					],Reply_params)
	),
	messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
	format("... ERROR composed~n",[]).


%
% handling different situations of ASK_ALL
%
handle_ask_all(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, and I can handle it
	%
	messaging:get_argument(Params,[	reply_with:Reply_with,		%get reply_with
					content:[Vars:Types,Query]]),		%get vars & query CIF
	process_ccql(ask_all,Vars,Types,Query,Result),
	messaging:set_argument([sender:Me,
				receiver:Sender,
				in_reply_to:Reply_with,
				content:Result
				],
				Reply_params),				%use as in_reply_to
	messaging:compose_ccql_core(tell,Reply_params,Reply_core).	%compose TELL core


handle_ask_all(Sender,Me,Params,none,Reply_core) :-
	%
	% no error, but problem in processing ICode
	%
	messaging:get_argument(Params,reply_with:Reply_with),		%get reply_with
	messaging:set_argument([sender:Me,
				receiver:Sender,
				in_reply_to:Reply_with,
				content:'problem in processing query'
				],
				Reply_params),
	messaging:compose_ccql_core(error,Reply_params,Reply_core).


handle_ask_all(Sender,Me,Params,Error,Reply_core) :-
	%
	% has error
	%
	(	(	messaging:get_argument(Params,[reply_with:Reply_with]),
			!,
			messaging:set_argument([sender:Me,
						receiver:Sender,
						in_reply_with:Reply_with,
						language_cif,
						ontology:shared,
						content:Error
						],Reply_params)
		);
		messaging:set_argument([sender:Me,
					receiver:Sender,
					in_reply_with:Reply_with,
					language_cif,
					ontology:shared,
					content:Error
					],Reply_params)
	),
	messaging:compose_ccql_core(error,Reply_params,Reply_core),	%compose error core
	format("... ERROR composed~n",[]).



%
% process Icode and generate result in an explicit set
%
process_ccql(ask_one,Variables_wanted,Types,Cif,[Variables_wanted:Types,Result]) :-
	format("*** processing CCQL~n",[]),
	cif2i:cif2i(Cif,Icode),
	make_code(Icode,Symbol_table,Prolog_code),			%generate Prolog code
	!,
	match_symbols(Variables_wanted,Symbol_table,Prolog_vars),	%match variables against symbol table
	format("... finding one solution~n",[]),
	(       (       call(Prolog_code),
			Result=[Prolog_vars],
			format("=== one solution found~n",[])
		);
		(	Result=[],
			format("=== no solution found~n",[])
		)
	),
	!.

process_ccql(ask_all,Variables_wanted,Types,Cif,[Variables_wanted:Types,Result]) :-
	format("*** processing CCQL~n",[]),
	cif2i:cif2i(Cif,Icode),
	make_code(Icode,Symbol_table,Prolog_code),			%generate Prolog code
	!,
	format("...Prolog code...~n",[]),
	write(Prolog_code),
	nl,
	match_symbols(Variables_wanted,Symbol_table,Prolog_vars),	%match variables against symbol table
	format("... finding all solutions~n",[]),
	(	(	setof(Prolog_vars,Prolog_code,Result),		%find set-of
			format("=== all solutions found~n",[])
		);
		(	Result=[],
			format("=== no solutions found~n",[])
		)
	),
	!.

process_ccql(_,_,_,_,_) :-
        format("*** ERROR in processing CCQL~n",[]).




%
% unify with variables in symbol-table
%

match_symbols([],_,[]).

match_symbols([V|Vs],Symbol_table,[P|Ps]) :-
	match_single_symbol(V,Symbol_table,P),
	match_symbols(Vs,Symbol_table,Ps).

match_single_symbol(_,Symbol_table,_) :-
	var(Symbol_table),
	!,
	fail.

match_single_symbol(V,[S|_],P) :-
	S = (V=P),
	!.

match_single_symbol(V,[_|Ss],P) :-
	match_single_symbol(V,Ss,P).



%
% starts after loading
%

:-	format("*** run me with: go('wrapper.config',HOST,PORT)~n",[]).

%:- go.

