%
% CIF<->RDF mapping module
%
% The mapping process is driven by a RDFS file/resource that defines the CIF language.
% The user/programmer has to provide another file that maps the Prolog respresentations
% of a CIF & its properties.
%
% Using the same technique, we can map queries & result tuples.
%

:- module(c2t,[	c2t/3,
		c2t/5,
		t2c/3
		]).

:- [rdf2triple].				%the RDF<->triple mapping module
:- [triple].					%the triple store
:- ['../../prolog/pillow/pillow'].		%the Pillow XML<->tree mapping module
:- ['../../prolog/util/cif_util'].
:- ['../../prolog/util/prolog_detect'].
:- ['../../prolog/util/concat_atoms'].

%:- [class_property].

%
% load required library depending on the Prolog system in use
%
:- prolog_detect:prolog(Prolog),
	(	(	Prolog=sicstus,
			use_module(library(lists))
		);
		(	Prolog=pfdm,
			use_module(library(lists))
		);
		true
	).

%
% how do I call these in the final RDF document
% i.e. the names in their namespace definitions
%
name_in_rdf_doc(rdf,rdf).
name_in_rdf_doc(rdfs,rdfs).
name_in_rdf_doc(cif,cif).

%
% known URLs of CIF definition in RDFS
%
known_url(rdf,'http://www.w3.org/1999/02/22-rdf-syntax-ns').
known_url(rdfs,'http://www.w3.org/2000/01/rdf-schema').
known_url(compact_cif,'http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-compact-rdfs.xml').
known_url(classic_cif,'http://www.csd.abdn.ac.uk/~khui/akt/cif/cif-classic-rdfs.xml').

%
% known compiled RDF(S) files
%
known_file(compact_cif,'/home/staff/khui/kraft/prolog/rdf2triple/cif-compact-rdfs.triple').
known_file(classic_cif,'/home/staff/khui/kraft/prolog/rdf2triple/cif-classic-rdfs.triple').

%
% map representation to name
%
rep_to_name_map(classic,classic_cif).	%map representation to RDFS name
rep_to_name_map(compact,compact_cif).

%
%
%
%:- dynamic(rdf_content/2).	%storing RDF document content at a certain URL


%
% load the CIF RDFS file if needed
%

%
% no need to load
%
load_rdfs_if_needed(Name) :-
	known_url(Name,Url),		%find URL from name
	triple:rdf_content(Url,_),		%RDFS document already loaded
	!.

%
% try to load compiled local file
%
load_rdfs_if_needed(Name) :-
	known_file(Name,File),		%see if there is a compiled version in triple
	write('load CIF RDFS from local file '),
	write(File),
	write('...'),
	nl,
	see(File),
	read_all_triples,		%read triples
	seen,
	!.
%
% load RDFS & parse into triples
%
load_rdfs_if_needed(Url) :-
	write('loading '),
	write(Url),
	write('...'),
	nl,
        get_xml_tree(Url,Xml_tree),		%read & get XML tree from URL
	get_rdf_tree(Xml_tree,Rdf_tree),	%get RDF subtree from XML tree
        rdf2triple:rdf2triple(Url,Rdf_tree).	%parse Xml tree as triple
%	move_triple_to_local(Url).		%move triples to local space

read_all_triples :-
	repeat,
	read(Something),
	read_all_triples2(Something).

read_all_triples2(end_of_file) :-
	!.
read_all_triples2(Something) :-
	assertz(Something),
	fail.				%fail to loop


%
% read an XML tree from a URL
%
get_xml_tree(Url,Xml_tree) :-
        pillow:url_info(Url,Url_as_term),                               %convert URL into a term
        pillow:fetch_url(Url_as_term,[],Response),                      %fetch URL
        member(status(success,_,_),Response),                           %check for success
        member(content(Content_string),Response),
        pillow:xml2terms(Content_string,Xml_tree).                      %parse into XML tree

%
% get RDF subtree from an XML tree
%
get_rdf_tree(Xml_tree,Rdf_tree) :-
        Rdf_tree=elem('rdf:RDF',_),
        member(Rdf_tree,Xml_tree).

get_rdf_tree(Xml_tree,Rdf_tree) :-
        Rdf_tree=env('rdf:RDF',_,_),
        member(Rdf_tree,Xml_tree).



%
% map a CIF into a constraint represented as a list of triples
%
% +To_map		=	the CIF to map
% +Params		=	a list of parameters
% -Constraint_id	=	RDF id of the starting constraint resource
% -Triples		=	the list of triples
%
c2t(To_map,Params,Constraint_id,Triples) :-
	c2t(To_map,constraint,Params,Constraint_id,Triples).



%
% try to map a CIF to an instance of Class
%
% +To_map	=	the CIF to map
% +Class	=	map CIF into an instance of Class
%			Class is just a plain name in the CIF definition w/ no namespace specified
% +Params	=	a list in the form X=Y, where X=Y can be:
%			representation=classic/compact
% -Value	=	the mapped resource or value
% -All_triples	=	the mapped list of triples + namespace definitions
%
c2t(To_map,Class,Params,Value,All_triples) :-
	%
	% how do I call these namespaces in the final RDF doc
	%
	name_in_rdf_doc(rdf,My_rdf_ns),
	name_in_rdf_doc(rdfs,My_rdfs_ns),
	name_in_rdf_doc(cif,My_cif_ns),
	%
	% from representation (e.g. classic/compact), find out what URL to load CIF definition
	%
	member(representation=Rep,Params),		%get representation
	rep_to_name_map(Rep,Cif_name),			%find the name of CIF RDFS to load
	load_rdfs_if_needed(Cif_name),			%load CIF RDFS file locally or through HTTP
	%
	% compose cif namespace
	%
	known_url(Cif_name,Cif_rdfs_url),						%get URL of CIF RDFS from name
	concat_atoms:concat_atoms([Cif_rdfs_url,'#'],Cif_url_with_hash),		%append '#' to the end of URL
	Cif_namespace_def=namespace(My_cif_ns,Cif_url_with_hash),				%cif namespace definition
	%
	% compose rdf namespace
	%
	known_url(rdf,Rdf_url),
	concat_atoms:concat_atoms([Rdf_url,'#'],Rdf_url_with_hash),			%append '#' to the end of URL
	Rdf_namespace_def=namespace(My_rdf_ns,Rdf_url_with_hash),			%RDF namespace definition
	%
	% compose rdfs namespace
	%
	known_url(rdfs,Rdfs_url),
	concat_atoms:concat_atoms([Rdfs_url,'#'],Rdfs_url_with_hash),			%append '#' to the end of URL
	Rdfs_namespace_def=namespace(My_rdfs_ns,Rdfs_url_with_hash),				%RDFS namespace definition
	Namespace_defs=[Rdf_namespace_def,Rdfs_namespace_def,Cif_namespace_def],
	%
	% now map CIF to target class
	%
	c2t_class(Namespace_defs,To_map,My_cif_ns:Class,Params,[],Value,Triples),	%map CIF to Class, where Url is the CIF RDFS
	append(Namespace_defs,Triples,All_triples).
%	All_triples=[Rdf_namespace_def,Rdfs_namespace_def,Cif_namespace_def|Triples].
%	All_triples=[Namespace_def|Triples].
%	c2t_class(Cif_namespace,Cif_rdfs_url,Cif,id(Class),Params,[],Value,Triples).	%map CIF to Class, where Url is the CIF RDFS


%
% map a CIF into a class
%
% +Namespaces	=	a list of namespace/2 definitions
% +To_map	=	the CIF to map, in Prolog terms
% +Class	=	the quantified target CIF class in name:class format, e.g. cif:constraint
%			if this is a list, it will try mapping CIF to 1 of these classes
% +Params	=	the list of parameters in x=y format
% +History	=	the list of history in map/3 format
%			this helps to prevent going into an infinite loop
% -Value	=	the final mapped value of this CIF
%			this can be a resource ID or a atomics/numeric value
% -Triples	=	the result list of triples
%

%
% try mapping Cif into 1 of a list of classes
%
c2t_class(_,_,[],_,_,_,_) :-		%empty list, fail!
	!,
	fail.

%
% Cif is a singleton list
%
c2t_class(Namespaces,[To_map],Class,Params,History,Value,Triples) :-
	c2t_class(Namespaces,To_map,Class,Params,History,Value,Triples),
	!.

c2t_class(Namespaces,To_map,[C|_],Params,History,Value,Triples) :-
	c2t_class(Namespaces,To_map,C,Params,History,Value,Triples),	%Cif can be mapped into C
	!.

c2t_class(Namespaces,To_map,[_|Cs],Params,History,Value,Triples) :-
	c2t_class(Namespaces,To_map,Cs,Params,History,Value,Triples).	%try mapping Cif into 1 of the list Cs

%
% map Cif into an instance of Class
%
c2t_class(_,To_map,Class,_,History,_,_) :-
	member(map(To_map,Class),History),		%mapping already in history
	!,
	fail.					%fail!

%
% Cif is an atom, mapped into a RDFS Literal
%
c2t_class(Namespaces,To_map,Class,_,_,To_map,[]) :-
	(	atom(To_map);
		number(To_map)
	),
	atom(Class),
	find_namespace_and_url(Namespaces,rdfs,_,_,Rdfs_url_with_hash),
	concat_atoms([Rdfs_url_with_hash,'Literal'],Class),		%using the RDFS URL + Literal as an atom
	!.
%
% Cif is an atom, where target class is rdfs:Literal
%
c2t_class(Namespaces,To_map,Rdfs_ns_in_cif:'Literal',_,_,To_map,[]) :-
	%
	% find the URL of RDFS & CIF
	%
	find_namespace_and_url(Namespaces,rdfs,_,Rdfs_url,_),
	find_namespace_and_url(Namespaces,cif,_,Cif_url,_),
	%
	% now check if the name of RDF match in the CIF definition RDFS doc
	%
	triple:rdf_content(Cif_url,namespace(Rdfs_ns_in_cif,Rdfs_url)),		%find RDFS namespace in CIF RDFS definition
	!.



%
% try mapping Cif into an instance of a subclass of "Class" 1st
% this allows mapping to be "as specialised as possible"
%
c2t_class(Namespaces,To_map,My_cif_name:Class,Params,History,Value,Triples) :-
	%
	% get URLs of CIF & RDFS
	%
	find_namespace_and_url(Namespaces,cif,My_cif_name,Cif_url,_),
	find_namespace_and_url(Namespaces,rdfs,_,_,Rdfs_url_with_hash),
	triple:rdf_content(Cif_url,namespace(Rdfs_name_in_cif,Rdfs_url_with_hash)),	%get name of RDFS namespace name in CIF def RDFS doc
	%
	% find all subclasses of Class and try mapping Cif to 1 of them 1st
	%
	Template=My_cif_name:Subclass,
	Goal=triple:rdf_content(Cif_url,triple(id(Subclass),Rdfs_name_in_cif:subClassOf,id(Class))),
	findall(Template,call(Goal),All_subclasses),
	c2t_class(Namespaces,To_map,All_subclasses,Params,[map(To_map,My_cif_name:Class)|History],Value,Triples),	%try mapping Cif to 1 of these subclasses
	!.

%
% try mapping Cif into a resource of Class defined in the CIF RDFS
% we have to find & map all direct & inherited properties
%
c2t_class(Namespaces,To_map,My_cif_name:Class,Params,History,Value,Triples) :-
	%
	% find the name of RDFS in the CIF RDFS document
	%
	find_namespace_and_url(Namespaces,rdf,My_rdf_name,_,_),
	find_namespace_and_url(Namespaces,rdfs,_My_rdfs_name,_Rdfs_url,Rdfs_url_with_hash),
	find_namespace_and_url(Namespaces,cif,My_cif_name,Cif_url,_),
	%
	% find out how RDFS is called in the CIF def RDFS doc
	%
	triple:rdf_content(Cif_url,namespace(Rdfs_name_in_cif,Rdfs_url_with_hash)),	%see what RDFS is called in the CIF RDFS
	%
	% find all properties of this class
	%
	member(domain_rdfs=Domain_url,Params),					%get RDFS definition of domain model
	Goal1=(	%
		% find property of this class in the CIF RDFS definition
		% and find corresponding Prolog term structures from Cif
		%
		triple:rdf_content(Cif_url,triple(id(Property1),Rdfs_name_in_cif:domain,id(Class))),		%Property is on the Class
		triple:rdf_content(Cif_url,triple(id(Property1),Rdfs_name_in_cif:range,Property_class_term1)),	%and of the type Property_class
		quantify_class(Property_class_term1,My_cif_name,Quantified_property_class1),
		class_property(Class,Property1,Domain_url,To_map,Pterm1)				%find Prolog term of this property
		),
	Template1=property_to_map(My_cif_name:Class,My_cif_name:Property1,Pterm1,Quantified_property_class1),
	findall(Template1,call(Goal1),Local_properties),						%find all properties & their classes
	%
	% now find all inherited properties of this class
	%
	Goal2=(	%
		% find all inherited properties
		%
		find_superclass(Cif_url,Rdfs_name_in_cif,Class,Superclass),
		triple:rdf_content(Cif_url,triple(id(Property2),Rdfs_name_in_cif:domain,id(Superclass))),	%Property is on the Class
		triple:rdf_content(Cif_url,triple(id(Property2),Rdfs_name_in_cif:range,Property_class_term2)),	%and of the type Property_class
		quantify_class(Property_class_term2,My_cif_name,Quantified_property_class2),
		class_property(Class,Property2,Domain_url,To_map,Pterm2)				%find Prolog term of this property
		),
	Template2=property_to_map(My_cif_name:Class,My_cif_name:Property2,Pterm2,Quantified_property_class2),
	findall(Template2,call(Goal2),Inherited_properties),					%find all inherited properties & their classes
	%
	% now try to map all properties found
	%
	append(Local_properties,Inherited_properties,All_properties),				%append 2 lists of property together
	All_properties=[_|_],
	get_resource_id(Params,Value,New_params),						%get ID for this resource
	c2t_all_properties(Namespaces,Value,All_properties,New_params,[map(To_map,My_cif_name:Class)|History],Property_triples),	%map all properties to triples
	New_triple=triple(Value,My_rdf_name:type,My_cif_name:Class),
	Triples=[New_triple|Property_triples],
	!.

%
% map a list of properties, given their classes
%
c2t_all_properties(_,_,[],_,_,[]).			%end of property list

c2t_all_properties(Namespaces,Resource,[T|Ts],Params,History,Triples) :-
	T=property_to_map(_,Property,Pterm,Property_class),
	c2t_class(Namespaces,Pterm,Property_class,Params,History,Property_value,P_triples),
	New_triple=triple(Resource,Property,Property_value),
	c2t_all_properties(Namespaces,Resource,Ts,Params,History,Ps_triples),
	append([New_triple|P_triples],Ps_triples,Triples).

%
% generate a resource id
%
get_resource_id(Params,anonymous(Id),Params) :-
	member(constraint_id=generate,Params),
	!,
	cif_util:generate_counter(resource,Id).

get_resource_id(Params,id(Id),New_params) :-
	member(constraint_id=Id_list,Params),
	Id_list=[Id|Ids],				%Id_list is a list
	!,
	delete(Params,constraint_id=Id_list,Temp_params),
	New_params=[constraint_id=Ids|Temp_params].

get_resource_id(Params,anonymous(Id),Params) :-
	cif_util:generate_counter(resource,Id).


%
%
%
quantify_class(id(Class),Namespace,Namespace:Class) :-
	!.
quantify_class(Class,_,Class).

/*
get_constraint_id(Params,Id,Params) :-
        member(constraint_id=generate,Params),                                  %user wants to generate ID
        cif_util:generate_counter(constraint,'constr',Id),                      %generate constraint ID
        !.
%
% user provides a list of ID
%
get_constraint_id(Params,Id,New_params) :-
        member(constraint_id=Id_list,Params),                   %user supplies a list of ID
        Id_list=[Id|Ids],                                       %take 1st ID
        delete(Params,constraint_id=Id_list,Temp),              %remove constraint_id param from list
        New_params=[constraint_id:Ids|Temp],
        New_known=[id(constraint,Rdf_id)|Known],                %1 more constraint known
        delete(Params,constraint_id=Id_list,New_params),        %remove constraint_id param from list
        !.
%
% no ID to be used
%
get_constraint_id(Params,Id,Parms) :-
        var(Attribs),                                                           %CIF -> XML
        Attribs=[].
*/

%
% find superclass of a class
%
find_superclass(Cif_url,Rdfs_namespace,Cif_ns:Subclass,Cif_ns:Superclass) :-
	triple:rdf_content(Cif_url,triple(id(Subclass),Rdfs_namespace:subClassOf,id(Superclass))).

find_superclass(Cif_url,Rdfs_namespace,Cif_ns:Subclass,Cif_ns:Superclass) :-
	triple:rdf_content(Cif_url,triple(id(Subclass),Rdfs_namespace:subClassOf,id(Intermediate))),
	find_superclass(Cif_url,Rdfs_namespace,Cif_ns:Intermediate,Cif_ns:Superclass).


%
% find the name, URL w/ & w/o hash from a list of namespace definitions
%
find_namespace_and_url(Namespaces,What,Name_in_rdf_doc,Url,Url_with_hash) :-
	name_in_rdf_doc(What,Name_in_rdf_doc),				%get name in RDF doc
	member(namespace(Name_in_rdf_doc,Url_with_hash),Namespaces),	%find namespace def, getting URL w/ hash
	concat_atoms:add_prefix(Url,'#',Url_with_hash).			%remove hash


%
% starting from a resource, map a set of triples back to the CIF Prolog form
% triple/3 & namespace/2 must already be asserted as rdf_content/2
%
% +Doc	=	URL of RDF document from which we got the triples
% +Resource	=	RDF ID of the resource to start w/
% -Cif		=	the resulting Prolog term structure
%
t2c(Doc,Resource,Cif) :-
	%
	% get names of RDF & RDFS in this RDF doc
	%
	known_url(rdf,Rdf_url),							%get the known URL of RDF
	concat_atoms:add_prefix(Rdf_url,'#',Rdf_url_with_hash),			%append the '#'
	triple:rdf_content(Doc,namespace(Rdf_name_in_doc,Rdf_url_with_hash)),		%get its name in this RDF doc
	known_url(rdfs,Rdfs_url),						%get the known URL of RDFS
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),		%append the '#'
	triple:rdf_content(Doc,namespace(Rdfs_name_in_doc,Rdfs_url_with_hash)),	%get its name in this RDF doc
	%
	% find out which CIF RDFS to load
	% 1st we find out the name of the CIF namespace in this RDF doc
	% then we look for its URL
	%
	triple:rdf_content(Doc,triple(Resource,Rdf_name_in_doc:type,Cif_name_in_doc:_)),	%get the name of CIF namespace in this doc
	triple:rdf_content(Doc,namespace(Cif_name_in_doc,Cif_url_with_hash)),			%get the URL of the CIF RDFS definition
	concat_atoms:add_prefix(Cif_url,'#',Cif_url_with_hash),		%remove the '#' to get the URL
	known_url(Cif_name,Cif_url),					%find its name from the URL
	load_rdfs_if_needed(Cif_name),					%load CIF RDFS file locally or through HTTP
	%
	% we have to keep track of the namespace definitions
	% of RDF, RDFS & CIF
	% ***Notice this is a namespace/3
	%
	Rdf_namespace_def=namespace(rdf,Rdf_name_in_doc,Rdf_url_with_hash),
	Rdfs_namespace_def=namespace(rdfs,Rdfs_name_in_doc,Rdfs_url_with_hash),
	Cif_namespace_def=namespace(cif,Cif_name_in_doc,Cif_url_with_hash),
	Namespace_defs=[Rdf_namespace_def,Rdfs_namespace_def,Cif_namespace_def],
	%
	% find the class of this resource
	%
	triple:rdf_content(Doc,triple(Resource,Rdf_name_in_doc:type,Cif_name_in_doc:Class)),
	t2c(Namespace_defs,Doc,Resource,Cif_name_in_doc:Class,Cif).

%
% map a resource of a certain class into CIF
%
% +Namespaces	=	a list of namespace definition in namespace/3
% +Doc		=	the URL of this doc
% +Resource	=	the resource to map
% +Class	=	the class of this resource
% -Cif		=	the CIF in Prolog form
%

%
% resource is a RDFS Literal
%
t2c(Namespaces,Doc,Resource,Class,Resource) :-
	atom(Class),							%the class of this resource is an atom
	member(namespace(rdfs,_,Rdfs_url_with_hash),Namespaces),	%get RDFS URL in this doc
	add_prefix(Rdfs_url_with_hash,'Literal',Class),			%it is an RDFS Literal
	(	atom(Resource);						%make sure that Resource is either a string or number
		number(Resource)
	),
	!.

%
% resource is a class defined in the CIF RDFS
%
t2c(Namespaces,Doc,Resource,Cif_ns:Class,Cif) :-
	member(namespace(rdf,Rdf_ns,_),Namespaces),				%get the name of RDF in this doc
	member(namespace(rdfs,Rdfs_ns,_),Namespaces),				%get the name of RDF in this doc
	member(namespace(cif,Cif_ns,Cif_url_with_hash),Namespaces),
	add_prefix(Cif_url,'#',Cif_url_with_hash),
	%
	% find all properties, property values & their corresponding classes of this resource
	%
	pig,
	Goal=(	%
		% here we find all properties (instances) of this resource
		% 1st we check a property that exist in the current RDF doc
		% then we look into the CIF RDFS, making sure that this property has a domain of this class
		% or a subclass (as subclasses inherit properties from their superclass)
		% after that, we get the type (class) of the property
		%
		triple:rdf_content(Doc,triple(Resource,Cif_ns:Property,Property_value)),	%get property of this resource in this doc
		triple:rdf_content(Cif_url,triple(id(Property),Rdfs_ns:domain,id(Certain_class))),	%property domain is a certain class
		same_class_or_subclass(Cif_url,Cif_ns,Rdfs_ns,Certain_class,Class),
		triple:rdf_content(Cif_url,triple(id(Property),Rdfs_ns:range,Property_type)),		%get type of property
		%
		% the Property class can be a CIF class or RDFS Literal
		%
		(	(	Property_type=id(Property_class),				%property is a CIF class
				Property_class_with_namespace=Cif_ns:Property_class
			);
			(	atom(Property_type),						%or atomic
				Property_class_with_namespace=Property_type			%likely to be a RDFS Literal
			)
		)
	),
	Template=property_to_map(Cif_ns:Certain_class,Cif_ns:Property,Property_value,Property_class_with_namespace),
	findall(Template,call(Goal),All_properties),
	t2c_properties(Namespaces,Doc,Cif,All_properties).

t2c_properties(_,_,_,[]).

/*
t2c_properties(Namespaces,Doc,Cif,[P|Ps]) :-
	P=property_to_map(Cif_ns:Class,Cif_ns:Property,Property_value,Property_class_with_namespace),
%	member(namespace(rdfs,_,Rdfs_url),Namespaces),		%get RDFS URL from the CIF RDFS file
%	add_prefix(Rdfs_url,'#Literal',Property_class),
%	(	atom(Property_value);
%		number(Property_value)
%	),
	class_property(Class,Property,_,Cif,Property_value),		%get Prolog term of resource
	!,
	t2c_properties(Namespaces,Doc,Cif,Ps).
*/

t2c_properties(Namespaces,Doc,Cif,[P|Ps]) :-
	P=property_to_map(Cif_ns:Class,Cif_ns:Property,Property_value,Property_class_with_namespace),
	t2c(Namespaces,Doc,Property_value,Property_class_with_namespace,Pterm),	%get property Prolog term
	class_property(Class,Property,_,Cif,Pterm),			%partially instantiate CIF term by this property
	t2c_properties(Namespaces,Doc,Cif,Ps).

%
%
%
same_class_or_subclass(_,_,_,Class,Class) :-
	!.
same_class_or_subclass(Cif_doc,Cif_ns,Rdfs_ns,Subclass,Superclass) :-
	find_superclass(Cif_doc,Rdfs_ns,Cif_ns:Subclass,Cif_ns:Superclass).
pig.

