%
% A general mapping module between Prolog & RDF.
% Work for CIF constraint, query & result tuples.
%
% For each case, we need an RDFS file & a "class-property" file.
% The RDFS file defines the classes & properties while the "class-property" file
% tell us how to get the Prolog term that represent a property from the class,
% and vice versa.
%

:- module(prolog2triple,[	prolog2triple/6,	%map a Prolog term to RDF triples
				triple2prolog/3		%map a resource in a RDF document (URL) to a Prolog term
			]).

%:- dynamic(known/3).

:- [prolog2triple_config].			%load the config file
:- [rdf2triple].				%the RDF<->triple mapping module, need to parse RDF file into triples
:- [triple].					%the triple store module template
:- ['../../prolog/pillow/pillow'].		%the Pillow XML<->tree mapping module
:- ['../../prolog/util/cif_util'].		%CIF utilities
:- ['../../prolog/util/concat_atoms'].		%util predicate that concatenates atoms
:- ['../../prolog/util/prolog_detect'].		%Prolog platform detecting module

%
% load required library depending on the Prolog system in use
%
:- prolog_detect:prolog(Prolog),
	(	(	Prolog=sicstus,
			use_module(library(lists))
		);
		(	Prolog=pfdm,
			use_module(library(lists))
		);
		true					%SWI & ECLIPSe need not to load anything extra
	).

%
% we have to keep track of which file is loaded
% e.g. the class_property file & RDFS
%
% load/2 takes the form "load(What,Context)".
% "What" is either the atom "rdfs" or "class_property",
% meaning that the "RDFS" or "class-property file"
% of the "Context" is already loaded.
%
:- dynamic(loaded/2).

%
% Load a context.
% This usually includes the RDFS definition (in RDFS or compiled triple form)
% and the class-property file.
%
% +Context	=	the context to load
%			as defined in the configuration file
%
load_context(Context) :-
	known(Context,load,no),				%context not loadable
	!.

load_context(Context) :-
	known(Context,load,yes),			%context loadable
	load_rdfs_if_needed(Context),			%load RDFS
	load_class_property_file_if_needed(Context),	%load class-property file
	!.

load_context(Context) :-
	known(Context,load,yes),			%context loadable but fail
	write('*** ERROR: fail to load context: '),
	write(Context),
	nl,
	!,
	fail.

%
% load the CIF RDFS file if needed
%
% +Context	=	"Context" tells the predicate what to load
%			It will 1st look into the config. file to locate any RDFS or triple
%			defined for this context. If a triple file exist, it will be used 1st.
%			Otherwise it will try to load (& parse) an RDFS file.
%

%
% no need to load
%
load_rdfs_if_needed(Context) :-
	loaded(rdfs,Context),				%RDFS of this Context has already been loaded
	!.

%
% try to load compiled local triple file
%
load_rdfs_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable
	known(Context,triple_file,File),		%see if there is a compiled version in triple
	write('load CIF RDFS from local file '),
	write(File),
	write('...'),
	nl,
	!,
	load_rdfs_from_file(Context,File).

%
% load RDFS & parse into triples
%
load_rdfs_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable
	known(Context,url,Url),
	write('loading from URL: '),
	write(Url),
	write('...'),
	nl,
	!,
	load_rdfs_from_url(Context,Url).

load_rdfs_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable but fail
	write('*** ERROR: Cannot find RDFS file of context:'),
	write(Context),
	nl,
	!,
	fail.


%
% load RDFS from a local file
%
load_rdfs_from_file(Context,File) :-
	known(Context,load,yes),			%context is loadable
	see(File),
	read_all_triples,		%read triples
	seen,
	assertz(loaded(rdfs,Context)),	%another RDFS file loaded
	!.

load_rdfs_from_file(Context,File) :-
	known(Context,load,yes),
	write('*** ERROR loading RDFS from local file '),
	write(File),
	nl,
	!,
	fail.

%
% load RDFS from a URL using HTTP
%
load_rdfs_from_url(Context,Url) :-
        get_xml_tree(Url,Xml_tree),		%read & get XML tree from URL
	get_rdf_tree(Xml_tree,Rdf_tree),	%get RDF subtree from XML tree
        rdf2triple:rdf2triple(Url,Rdf_tree),	%parse Xml tree as triple
	assertz(loaded(rdfs,Context)),		%another RDFS file loaded
	!.

load_rdfs_from_url(_,Url) :-
	write('*** ERROR loading RDFS from URL: '),
	write(Url),
	nl,
	!,
	fail.

%
% read all triples from stdin
%
read_all_triples :-
	repeat,
	read(Something),
	read_all_triples2(Something).

read_all_triples2(end_of_file) :-
	!.
read_all_triples2(Something) :-
	assertz(Something),
	fail.				%fail to loop


%
% read an XML tree from a URL
%
% +Url		=	the URL to retrieve
% -Xml_tree	=	the retrieved XML tree (in Pillow's format)
%
get_xml_tree(Url,Xml_tree) :-
        pillow:url_info(Url,Url_as_term),                               %convert URL into a term
        pillow:fetch_url(Url_as_term,[],Response),                      %fetch URL
        member(status(success,_,_),Response),                           %check for success
        member(content(Content_string),Response),
        pillow:xml2terms(Content_string,Xml_tree).                      %parse into XML tree

%
% get RDF subtree from an XML tree
%
% +Xml_tree	=	the XML tree (in Pillow's format)
% -Rdf_tree	=	the RDF subtree in the XML tree
%
get_rdf_tree(Xml_tree,Rdf_tree) :-
        Rdf_tree=elem('rdf:RDF',_),
        member(Rdf_tree,Xml_tree).

get_rdf_tree(Xml_tree,Rdf_tree) :-
        Rdf_tree=env('rdf:RDF',_,_),
        member(Rdf_tree,Xml_tree).

%
% load class-property file
%
load_class_property_file_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable
	loaded(class_property,Context),			%file already loaded, no need to reload
	!.

load_class_property_file_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable
	known(Context,class_property,File),		%filename found
	!,
	load_class_property_file(Context,File).

load_class_property_file_if_needed(Context) :-
	known(Context,load,yes),			%context is loadable
	write('*** ERROR: Cannot find class-property file of context: '),
	write(Context),
	nl,
	fail.

%
% load/consult class-property file
%
load_class_property_file(Context,File) :-
	[File],						%load new one for this context
%	see(File),
%	read_all_triples,
%	seen,
	asserta(loaded(class_property,Context)).


%
% map a Prolog term to an instance of Class
%
% +Prolog	=	the Prolog term to map
% +Context	=	the "context" within which this mapping takes place
%			From the "Context", we determine which RDFS to load in order
%			to resolve class name.
%			*** NOTE:	Although the "Class" argument may contain an optional
%			***		namespace, the namespace still depends on the "Context", within
%			***		which the namespace is defined.
% +Class	=	the RDF class (as defined in this context RDFS) to map
%			"Class" may contain an optional namespace name
% +Params	=	a list in the form X=Y
%			this list of parameters will be passed
%			to the class_property/5 predicate defined in the class-property file of each context
% -Value	=	the mapped resource (an id/1) or value (an atom)
% -All_triples	=	the mapped list of triples + namespace definitions
%
prolog2triple(Prolog,Context,Class,Params,Value,All_triples) :-
	%
	% how do I call these namespaces in the final RDF doc
	%
	known(rdf,name_in_rdf,My_rdf_ns),
	known(rdfs,name_in_rdf,My_rdfs_ns),
	known(Context,name_in_rdf,Context_ns),
	%
	% load RDFS definition of this "Context"
	%
	load_rdfs_if_needed(Context),
	%
	% load class-property file
	%
	load_class_property_file_if_needed(Context),
	%
	% compose "Context" namespace
	%
	known(Context,url,Context_rdfs_url),						%get URL of "Context" RDFS from name
	concat_atoms:add_prefix(Context_rdfs_url,'#',Context_url_with_hash),		%add '#' to the end of URL
	Context_namespace_def=namespace(Context_ns,Context_url_with_hash),		%namespace definition
	%
	% compose rdf namespace
	%
	known(rdf,url,Rdf_url),
	concat_atoms:add_prefix(Rdf_url,'#',Rdf_url_with_hash),				%add '#' to the end of URL
	Rdf_namespace_def=namespace(My_rdf_ns,Rdf_url_with_hash),			%RDF namespace definition
	%
	% compose rdfs namespace
	%
	known(rdfs,url,Rdfs_url),
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),			%add '#' to the end of URL
	Rdfs_namespace_def=namespace(My_rdfs_ns,Rdfs_url_with_hash),			%RDFS namespace definition
	%
	% namespace definitions of RDF, RDFS & "Context"
	%
	Namespace_defs=[Rdf_namespace_def,Rdfs_namespace_def,Context_namespace_def],
	%
	% now map CIF to target class
	%
	prolog2triple_class(Prolog,Context,Class,Params,[],Value,Triples),		%map Prolog term to Class
	append(Namespace_defs,Triples,All_triples).					%add namespace defs to triple list


%
% map a Prolog term into an instance of a RDF class
%
% +Prolog	=	the Prolog "thing" to map
% +Context	=	the context in which the class is defined
% +Class	=	the target RDF class name, may contain an optional namespace
%			if this is a list, it will try mapping to 1 of these classes
% +Params	=	the list of parameters in x=y format
% +History	=	the list of history in map/3 format
%			this helps to prevent going into an infinite loop
% -Value	=	the final mapped value of this Prolog "thing"
%			this can be a resource ID or a atomics/numeric value
% -Triples	=	the result list of triples
%

%
% try mapping Cif into 1 of a list of classes
%
prolog2triple_class(_,_,[],_,_,_,_) :-		%empty list, fail!
	!,
	fail.

%
% Prolog term to map is a singleton list
%
prolog2triple_class([Prolog],Context,Class,Params,History,Value,Triples) :-
	prolog2triple_class(Prolog,Context,Class,Params,History,Value,Triples),
	!.

%
% Prolog term can be mapped to the 1st class in the list
%
prolog2triple_class(Prolog,Context,[C|_],Params,History,Value,Triples) :-
	prolog2triple_class(Prolog,Context,C,Params,History,Value,Triples),	%term can be mapped into class C
	!.

%
% Prolog term can be mapped to 1 of the classes in the tail
%
prolog2triple_class(Prolog,Context,[_|Cs],Params,History,Value,Triples) :-
	prolog2triple_class(Prolog,Context,Cs,Params,History,Value,Triples).	%try mapping Prolog term into 1 of the list Cs

%
% map Prolog into an instance of Class defined in Context
%
% +Prolog	=	the Prolog "thing" to map
% +Context	=	the context in which the result value/resource is defined
% +Class	=	the target class in context:class form
% +Params	=	a list of parameters in the form X=Y
% +History	=	the mapping history so far, in map/3 format
% -Value	=	the final mapped value of this Prolog "thing"
%			this can be a resource ID or a atomics/numeric value
% -Triples	=	the result list of triples
%
prolog2triple_class(Prolog,Context,Class,_,History,_,_) :-
	member(map(Prolog,Context,Class),History),		%mapping in history, tried this before
	!,
	fail.							%fail!

%
% Prolog term is an atom, mapped into a RDFS Literal
% in this case, Class is the atom '.....#Literal' that points to Literal class in RDFS
% so there is no namespace in the class name here
%

%
% map an atom or number to a Literal in the "rdfs" context
%
prolog2triple_class(Prolog,rdfs,'Literal',_,_,Prolog,[]) :-
	(	atom(Prolog);
		number(Prolog)
	),
	!.

prolog2triple_class(Prolog,_,Class,_,_,Prolog,[]) :-
	(	atom(Prolog);						%term to map is an atom
		number(Prolog)						%or a number
	),
	atom(Class),
	known(rdfs,url,Rdfs_url),					%find URL of RDFS
	concat_atoms:add_prefix(Rdfs_url,'#Literal',Class),		%Class is the atom that points to RDFS#Literal
	!.

%
% Prolog term to map is an atom, where target class is rdfs:Literal
%
prolog2triple_class(Prolog,Context,Rdfs_ns:'Literal',_,_,Prolog,[]) :-
	(	atom(Prolog);						%term to map is an atom
		number(Prolog)						%or a number
	),
	%
	% find out how RDFS is called in this Context
	%
	known(Context,url,Context_url),						%find context URL
	known(rdfs,url,Rdfs_url),						%find RDFS URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),		%append '#' to the end of RDFS URL
	triple:rdf_content(Context_url,namespace(Rdfs_ns,Rdfs_url_with_hash)),	%check that RDFS is called this name in this context
	!.

%
% try mapping Prolog term into an instance of a subclass of "Class" 1st
% this allows mapping to be "as specialised as possible"
%

%
% Class contains a namespace, resolve correct context 1st
%
prolog2triple_class(Prolog,Context,Namespace:Class,Params,History,Value,Triples) :-
	%
	% load RDFS & class-property of namespace if needed
	%
	known(Context,url,Context_url),								%get URL of Context
	triple:rdf_content(Context_url,namespace(Namespace,Namespace_url_with_hash)),		%get namespace definition in this Context
	concat_atoms:add_prefix(Namespace_url,'#',Namespace_url_with_hash),			%remove '#' to get URL
	known(Namespace_context,url,Namespace_url),						%get context of this Namespace
	load_context(Namespace_context),							%load namespace context
	prolog2triple_class(Prolog,Namespace_context,Class,Params,History,Value,Triples).	%switch to the correct context

%
% Class already in Context, try mapping to a subclass of Class
%
prolog2triple_class(Prolog,Context,Class,Params,History,Value,Triples) :-
	atom(Class),
	%
	% get URLs of CIF & RDFS
	%
	known(Context,url,Context_url),		%get Context URL
	known(rdfs,url,Rdfs_url),		%get RDFS URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),
	triple:rdf_content(Context_url,namespace(Rdfs_name,Rdfs_url_with_hash)),	%get name of RDFS namespace name in Context
	%
	% *** NOTE: here we find subclass in the same context only
	%
	Template=Subclass,
	Goal=triple:rdf_content(Context_url,triple(id(Subclass),Rdfs_name:subClassOf,id(Class))),
	findall(Template,call(Goal),All_subclasses),
	prolog2triple_class(Prolog,Context,All_subclasses,Params,[map(Prolog,Context,Class)|History],Value,Triples),	%try mapping Term to 1 of these subclasses
	!.

%
% try mapping term into a resource of Class defined in Context
% we have to find & map all direct & inherited properties
%
/*
prolog2triple_class(Prolog,Context,Class,Params,History,Value,Triples) :-
	atom(Class),
	%
	% find name of RDFS i this Context
	%
	known(Context,url,Context_url),							%find Context of this URL
	known(rdfs,url,Rdfs_url),							%get RDFS URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),			%append '#' to end of URL
	triple:rdf_content(Context_url,namespace(Rdfs_name,Rdfs_url_with_hash)),	%find RDFS name in this Context
	Goal=(	%
		% domain class match
		%
		triple:rdf_content(Context_url,triple(Property,Rdfs_name:domain,Domain_class)),
		find_context(Context_url,Context,Domain_class,Domain_class_context,Domain_class_local_name),
		load_context(Domain_class_context),
		same_class_or_subclass(Domain_class_context,Domain_class_local_name,Context,Class),
		%
		% now get range
		%
		triple:rdf_content(Context_url,triple(Property,Rdfs_name:range,Range_class)),
		find_context(Context_url,Context,Range_class,Range_class_context,Range_class_local_name),
		tidy_class_for_property(Property_class_term2,Property_class2),

		class_property(Context,Class,Property,Params,Prolog,Pterm)				%find Prolog term of this property
	),
	Template=property_to_map(Context,Class,Property,Pterm,Property_class),
	findall(Template1,call(Goal1),Local_properties),				%find all properties & their classes
*/

prolog2triple_class(Prolog,Context,Class,Params,History,Value,Triples) :-
	atom(Class),									%no namespace specified
	%
	% find the name of RDFS in this Context
	%
	known(Context,url,Context_url),							%find Context of this URL
	known(rdfs,url,Rdfs_url),							%get RDFS URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),			%append '#' to end of URL
	triple:rdf_content(Context_url,namespace(Rdfs_name,Rdfs_url_with_hash)),	%find RDFS name in this Context
	%
	% find all properties of this class
	%
	Goal1=(	%
		% find property of this class in this Context
		% and find corresponding Prolog term structures of the property
		%
		triple:rdf_content(Context_url,triple(id(Property1),Rdfs_name:domain,id(Class))),		%Property1 is on the Class
		triple:rdf_content(Context_url,triple(id(Property1),Rdfs_name:range,Property_class_term1)),	%and of the type Property_class_term1
		tidy_class_for_property(Property_class_term1,Property_class1),
		class_property(Context,Class,Property1,Params,Prolog,Pterm1)			%find property as a Prolog term
		),
	%
	% this is the thing we get in the list, to be mapped later
	%
	Template1=property_to_map(Context,Class,Property1,Pterm1,Property_class1),
	findall(Template1,call(Goal1),Local_properties),				%find all properties & their classes
	%
	% now find all inherited properties of this class
	%
	Goal2=(	%
		% find all inherited properties
		%
		find_superclass(Context_url,Rdfs_name,Class,Superclass),
		triple:rdf_content(Context_url,triple(id(Property2),Rdfs_name:domain,id(Superclass))),	%Property is on the Class
		triple:rdf_content(Context_url,triple(id(Property2),Rdfs_name:range,Property_class_term2)),	%and of the type Property_class
		tidy_class_for_property(Property_class_term2,Property_class2),
		class_property(Context,Class,Property2,Params,Prolog,Pterm2)				%find Prolog term of this property
		),
	%
	% the things we need to map later
	%
	Template2=property_to_map(Context,Class,Property2,Pterm2,Property_class2),
	findall(Template2,call(Goal2),Inherited_properties),					%find all inherited properties & their classes
	%
	% now try to map all properties found
	%
	append(Local_properties,Inherited_properties,All_properties),				%append 2 lists of property together
	All_properties=[_|_],
	get_resource_id(Params,Value,New_params),						%get ID for this resource
	prolog2triple_all_properties(Value,All_properties,New_params,[map(Prolog,Context,Class)|History],Property_triples),	%map all properties to triples
	%
	% now make up this triple
	%
	known(Context,name_in_rdf,Context_name_in_final_doc),
	known(rdf,name_in_rdf,Rdf_name_in_final_doc),
	New_triple=triple(Value,Rdf_name_in_final_doc:type,Context_name_in_final_doc:Class),
	Triples=[New_triple|Property_triples],
	!.

%
% map a list of properties, given their classes
%
prolog2triple_all_properties(_,[],_,_,[]).			%end of property list

prolog2triple_all_properties(Resource,[T|Ts],Params,History,Triples) :-
	T=property_to_map(Context,_,Property,Pterm,Property_class),
	prolog2triple_class(Pterm,Context,Property_class,Params,History,Property_value,P_triples),
	%
	% get the namespace of this Context in the final RDF doc
	%
	known(Context,name_in_rdf,Context_name_in_rdf),
	New_triple=triple(Resource,Context_name_in_rdf:Property,Property_value),
	prolog2triple_all_properties(Resource,Ts,Params,History,Ps_triples),
	append([New_triple|P_triples],Ps_triples,Triples).

%
% generate a resource id
%
get_resource_id(Params,anonymous(Id),Params) :-
	member(constraint_id=generate,Params),
	!,
	cif_util:generate_counter(resource,Id).

get_resource_id(Params,id(Id),New_params) :-
	member(constraint_id=Id_list,Params),
	Id_list=[Id|Ids],				%Id_list is a list
	!,
	delete(Params,constraint_id=Id_list,Temp_params),
	New_params=[constraint_id=Ids|Temp_params].

get_resource_id(Params,anonymous(Id),Params) :-
	cif_util:generate_counter(resource,Id).


%
% if id(Class) is found, return Class
% we have to make sure that the id/1 is gone
%
tidy_class_for_property(id(Class),Class) :-
	!.
tidy_class_for_property(Class,Class).

quantify_class(id(Class),Namespace,Namespace:Class) :-
	!.
quantify_class(Class,_,Class).


%
% find superclass of a class
%
% +Url		=	URL of the RDFS document in which these classes are defined
% +Rdfs_name	=	the name of the RDFS namespace in the document
% +Subclass	=	the subclass, without id/1 or any namespace
% ?Superclass	=	the superclass, without id/1 or any namespace
%
find_superclass(Context_url,Rdfs_name,Subclass,Superclass) :-
	triple:rdf_content(Context_url,triple(id(Subclass),Rdfs_name:subClassOf,id(Superclass))).

find_superclass(Context_url,Rdfs_name,Subclass,Superclass) :-
	triple:rdf_content(Context_url,triple(id(Subclass),Rdfs_name:subClassOf,id(Intermediate))),
	find_superclass(Context_url,Rdfs_name,Intermediate,Superclass).


%
% starting from a resource, map a set of triples back to their Prolog form
% triple/3 & namespace/2 must already be asserted as rdf_content/2 in module "triple"
%
% +Doc_url	=	URL of RDF document from which we got the triples
% +Resource	=	RDF ID of the resource to start w/
% -Term		=	the resulting Prolog term structure
%
triple2prolog(Doc_url,Resource,Fixed) :-
	%
	% generate a new context for this Doc_url if needed
	%
	(	known(Default_context,url,Doc_url);
		(	cif_util:generate_counter(context,'context_',Default_context),
			assertz(known(Default_context,url,Doc_url))
		)
	),
	!,
	%
	% get names of RDF & RDFS in this RDF doc
	%
	known(rdf,url,Rdf_url),							%get the URL of RDF
	concat_atoms:add_prefix(Rdf_url,'#',Rdf_url_with_hash),			%append the '#'
	triple:rdf_content(Doc_url,namespace(Rdf_name,Rdf_url_with_hash)),	%get the name of RDF in this doc
	%
	% load context required by this Resource type
	%
	triple:rdf_content(Doc_url,triple(Resource,Rdf_name:type,Resource_class)),	%find Resource rdf:type
	find_context(Doc_url,Default_context,Resource_class,Class_context,Class_local_name),
	load_context(Class_context),
	%
	% we have to keep track of the namespace definitions
	% of RDF, RDFS & CIF
	% ***Notice this is a namespace/3
	%
	%
	% proceed
	%
	triple2prolog(Doc_url,Resource,Class_context,Class_local_name,Term),
	triple2prolog_roundup(Class_context,Term,Fixed),
	!.



%
% Resource is a RDFS Literal
%
triple2prolog(_,Resource,_,Class,Resource) :-
	atom(Class),
	known(rdfs,url,Rdfs_url),
	concat_atoms:add_prefix(Rdfs_url,'#Literal',Class),
	(	atom(Resource);
		number(Resource)
	),
	!.

triple2prolog(_,Resource,rdfs,'Literal',Resource) :-		%context is rdfs
	(	atom(Resource);
		number(Resource)
	),
	!.

%
% map Resource to a RDF class in Context
%
triple2prolog(Doc,Resource,Context,Class,Term) :-
	%
	% find all properties, property values & their corresponding classes of this resource
	%
	Goal=(	%
		% for each property of this resource, we have to find out the corresponding Prolog term to map
		%
		% 1.	we get a property of this resource in the current (RDF) Doc
		% 2.	optionally we may have to load in a context (if the property is defined elsewhere)
		%	(i.e. loading in the RDFS & class-property file)
		% 3.	find out the domain & range of this property
		%
		triple:rdf_content(Doc,triple(Resource,Property,Property_value)),		%get property of this resource
		find_context(Doc,Context,Property,Property_context,Property_local_name),	%find context of this property
		%
		% make sure that it is not rdf & rdfs "things" (e.g. rdf:type, etc.)
		%
		\+(Property_context=rdf),
		\+(Property_context=rdfs),
%		not(Property_context=rdf),
%		not(Property_context=rdfs),
		load_context(Property_context),						%load property context if needed
		known(Property_context,url,Property_context_url),			%get URL of RDF doc that contains the property context
		known(rdfs,url,Rdfs_url),
		concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),
		%
		% check if current class matches domain of this property
		%
		triple:rdf_content(Property_context_url,namespace(Rdfs_name_in_property_context,Rdfs_url_with_hash)),	%find RDFS name
		triple:rdf_content(Property_context_url,triple(id(Property_local_name),Rdfs_name_in_property_context:domain,Domain_class)),
		find_context(Property_context_url,Property_context,Domain_class,Domain_class_context,Domain_class_local_name),
		load_context(Domain_class_context),
		same_class_or_subclass(Domain_class_context,Domain_class_local_name,Context,Class),	%the classes match
		%
		% get range class of this property
		%
		triple:rdf_content(Property_context_url,triple(id(Property_local_name),Rdfs_name_in_property_context:range,Range_class)),
		find_context(Property_context_url,Property_context,Range_class,Range_class_context,Range_class_local_name),
		load_context(Range_class_context)
		%
		% check that the domain class match
		%
	),
	Template=property_to_map(Domain_class_context,Domain_class_local_name,Range_class_context,Range_class_local_name,Property_context,Property_local_name,Property_value),
	findall(Template,call(Goal),All_properties),
	triple2prolog_properties(Doc,All_properties,Term).


%
% map a list of properties
%
triple2prolog_properties(_,[],_).

triple2prolog_properties(Doc,[P|Ps],Term) :-
	P=property_to_map(Domain_context,Domain_class,Range_context,Range_class,_Property_context,Property,Property_value),
	triple2prolog(Doc,Property_value,Range_context,Range_class,Pterm),
	class_property(Domain_context,Domain_class,Property,_,Term,Pterm),		%partially instantiate CIF term by this property
	triple2prolog_properties(Doc,Ps,Term).



%
% find the context & local name of an item
%
% +Doc			=	URL of the document where this "Item" is found
% +Default_context	=	the default context
% +Item			=	the item, it can be in the form a id/1 or namespace:local_name
% -Context		=	context of the item
% -Local_name		=	the local name of the item
%

%
% locally defined
%
find_context(_,_,Something,Context,Local_name) :-
	atom(Something),
	known(Context,url,Context_url),
	concat_atoms:add_prefix(Context_url,'#',Context_url_with_hash),
	concat_atoms:add_prefix(Context_url_with_hash,Local_name,Something),
	!.

find_context(_,Default_context,id(Something),Default_context,Something).

%
% defined in another namespace
% context already known
%
find_context(Current_doc,_,Namespace:Local_name,Namespace_context,Local_name) :-
	triple:rdf_content(Current_doc,namespace(Namespace,Namespace_url_with_hash)),
	concat_atoms:add_prefix(Namespace_url,'#',Namespace_url_with_hash),
	known(Namespace_context,url,Namespace_url),
	!.

%
% defined in another namespace
% context not yet known, has to create
%
find_context(Current_doc,_,Namespace:Local_name,New_context,Local_name) :-
	triple:rdf_content(Current_doc,namespace(Namespace,Namespace_url_with_hash)),
	concat_atoms:add_prefix(Namespace_url,'#',Namespace_url_with_hash),
	cif_util:generate_counter(context,'context_',New_context),		%create a new context name
	assertz(known(New_context,url,Namespace_url)),
	!.


%
% see if class1 is a subclass of class2
%
% +Context1	=	context of class1
% +Context2	=	context of class2
% +Class1	=	class 1 w/o id/1 or any namespace
% +Class2	=	class 2 w/o id/1 or any namespace
%

%
% same class
%
same_class_or_subclass(Context,Class,Context,Class) :-
	!.

%
% classes in the same context
%
same_class_or_subclass(Context,Subclass,Context,Superclass) :-			%both classes in the same context
	known(Context,url,Doc_url),						%find URL of this context
	known(rdfs,url,Rdfs_url),						%find the URL of RDFS
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),		%add '#' to the end of URL
	triple:rdf_content(Doc_url,namespace(Rdfs_name,Rdfs_url_with_hash)),	%find name of RDFS namespace in this context
	find_superclass(Doc_url,Rdfs_name,Subclass,Superclass),			%Class1 is a subclass of Class2
	!.

%
% classes in different context
%

%
% in ConText1, it is mentioned that Class1 is a subclass of Class2 (using Class2 URL)
%
same_class_or_subclass(Context1,Class1,Context2,Class2) :-
	known(rdfs,url,Rdfs_url),						%get RDFS URL
	known(Context1,url,Context1_url),					%get Context1 URL
	known(Context2,url,Context2_url),					%get Context2 URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),		%add '#' to RDFS URL
	concat_atoms:add_prefix(Context2_url,'#',Context2_url_with_hash),	%add '#' to Context2 URL
	triple:rdf_content(Context1_url,namespace(Rdfs_name_in_context1,Rdfs_url_with_hash)),	%find RDFS name in Context1
	concat_atoms:add_prefix(Context2_url_with_hash,Class2,Class2_url),
	%
	% in Context1, Class1 is declared to be a subclass of Class2, using Class2 URL
	%
	triple:rdf_content(Context1_url,triple(id(Class1),Rdfs_name_in_context1:subClassOf,Class2_url)),
	!.

%
% same as above, but using namespace:Class2
%
same_class_or_subclass(Context1,Class1,Context2,Class2) :-
	known(rdfs,url,Rdfs_url),								%get RDFS URL
	known(Context1,url,Context1_url),							%get Context1 URL
	known(Context2,url,Context2_url),							%get Context2 URL
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),				%add '#' to RDFS URL
	concat_atoms:add_prefix(Context2_url,'#',Context2_url_with_hash),			%add '#' to Context2 URL
	triple:rdf_content(Context1_url,namespace(Rdfs_name_in_context1,Rdfs_url_with_hash)),	%find RDFS name in Context1
	triple:rdf_content(Context1_url,namespace(Namespace2,Context2_url_with_hash)),		%find namespace of Context2 in Context1
	%
	% in Context1, Class1 is declared to be a subclass of Class2, using namespace:Class2 style notation
	%
	triple:rdf_content(Context1_url,triple(id(Class1),Rdfs_name_in_context1:subClassOf,Namespace2:Class2)),
	!.

%
% Class1 is a subclass of a class in Context1,
% which in turn is a subclass of Class2 (in Context2)
%
same_class_or_subclass(Context1,Class1,Context2,Class2) :-
	known(Context1,url,Context1_url),							%find Context1 URL
	known(rdfs,url,Rdfs_url),								%find known URL of RDFS
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),				%add '#' to RDFS URL
	triple:rdf_content(Context1,namespace(Rdfs_name_in_context1,Rdfs_url_with_hash)),	%find RDFS name in Context1
	find_superclass(Context1_url,Rdfs_name_in_context1,Class1,Temp_class),			%find a superclass of Class1 in Context1
	same_class_or_subclass(Context1,Temp_class,Context2,Class2),				%which is a subclass of Class2
	!.

%
% Class2 is a subclass of a class in Context2,
% which in turn is a subclass of Class2 (in Context2)
%
same_class_or_subclass(Context1,Class1,Context2,Class2) :-
	known(Context2,url,Context2_url),							%find URL of Context2
	known(rdfs,url,Rdfs_url),								%find URL of RDFS
	concat_atoms:add_prefix(Rdfs_url,'#',Rdfs_url_with_hash),				%add '#' to RDFS URL
	triple:rdf_content(Context2,namespace(Rdfs_name_in_context2,Rdfs_url_with_hash)),	%find RDFS name in Context2
	find_superclass(Context2_url,Rdfs_name_in_context2,Temp_class,Class2),			%find a subclass of Class2 in Context2
	same_class_or_subclass(Context1,Class1,Context2,Temp_class),				%which is a superclass of Class1
	!.

