:- module(rdf2triple,[	rdf2triple/2,		%parse RDF into triples
			reset/0,		%reset counter, remove triples & namespace
			clean_triple/0,		%remove triples & namespace
			triple2rdf/2,		%map triples to RDF tree, no XML header
			triple2rdf/3		%map triples to RDF tree start from a certain resource, no XML header
			]).

:- [triple].				%use the triple module, where we store the generated triple/3 & namespace/2
:- ['../util/sort_precedence'].
:- ['../util/my_union'].
:- ['../util/concat_atoms'].
:- ['../util/prolog_detect'].
:- ['../util/parse_url'].		%parse URL module

:- prolog_detect:prolog(Prolog),
	(	(	Prolog=sicstus,
			use_module(library(lists)),
			['../util/flatten']
		);
		(	Prolog=pfdm,
			use_module(library(lists)),
			['../util/flatten']
		);
		true
	).

:- dynamic(counter/2).		%counter

%
% these properties are better as attributes
%
property_as_attribute([	rdf:type,
			rdf:'ID',
			rdf:about,
			rdf:resource
			]).


%
% reset all counters, remove namespace & triples
%
reset :-
	reset_counter(local_ns),
	reset_counter(anonymous_resource),
	clean_triple.

clean_triple :-
	retractall(triple:rdf_content(_,_)).
%	namespace(_,_)),
%	retractall(triple(_,_,_)).

%
% parse an XML tree, storing all triples as triple/3
% & name space definitions as namespace/2  inside the rdf_content/2 structure
% in the module "triple"
%
% +Doc		=	URL of RDF document
%			this is used to identify different RDF documents in the triple space
% +Xml_tree	=	Prolog terms representing an RDF doc as an XML tree
%
rdf2triple(Doc,Xml_tree) :-
	parse_rdf(Doc,rdf,Xml_tree).


%
% parse an XML tree starting w/ the element rdf:RDF
%
% RDF ::=	'<rdf:RDF namespaceDef* >' description* '</rdf:RDF>'
%
% +Doc		=	URL of RDF document to parse
% +Default_ns	=	the default namespace to use
% +Tree		=	the XML tree to parse
%

%
% single 'rdf:RDF' element
%
parse_rdf(Doc,Default_ns,Tree) :-
	Tree=elem(Tag,Attribs),
	%
	% extract optional local namespace def & one/more XML namespace def
	%
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),	%get any local namespace def in attribute list
	extract_namespace_definition(Doc,Attribs_temp,_),			%extract other namespace def from attribute list
	quantify_tag(Tag,Default_ns,Local_ns,rdf:'RDF').			%quantify the tag


%
% 'rdf:RDF' environment
%
parse_rdf(Doc,Default_ns,Tree) :-
	Tree=env(Tag,Attribs,Nested_list),		%there may be some subjects
	%
	% quantify this tag, make sure it is 'rdf:RDF'
	%
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),	%get any local namespace def in attributes list
	extract_namespace_definition(Doc,Attribs_temp,_),			%extract other namespace def from attribute list
	quantify_tag(Tag,Default_ns,Local_ns,rdf:'RDF'),			%quantify the tag
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	parse_description_zero_or_more(Doc,New_default_ns,Nested_list).		%parse nested elements




%
% parse zero of more rdf:Description in a list
%
% +Doc		=	URL of the RDF document
% +Default_ns	=	default namespace
% +List		=	a list of rdf:Description subtrees
%
parse_description_zero_or_more(_,_,[]).					%end of list

parse_description_zero_or_more(Doc,Default_ns,[T|Ts]) :-
	parse_description(Doc,Default_ns,T,_),				%parse 1st element
	!,
	parse_description_zero_or_more(Doc,Default_ns,Ts).		%parse remaining


%
% parse a single rdf:Description
%
% +Doc		=	URL of the RDF document
% +Default_ns	=	default namespace
% +Tree		=	subtree to be parsed as a rdf:Description
%
% description    ::= '<rdf:Description' idAboutAttr? propAttr* '/>'
%                    | '<rdf:Description' idAboutAttr? propAttr* '>' propertyElt* '</rdf:Description>'
%                    | resourceNode
%

%
% a comment, skip it
%
parse_description(_,_,T,_) :-
	T=comment(_),
	!.

%
% rdf:Description w/ no nested element
%
parse_description(Doc,Default_ns,T,Subject) :-
	T=elem(Tag,Attribs),								%element w/ no nested element
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),		%get any local namespace def
	extract_namespace_definition(Doc,Attribs_temp,Attribs_left),			%extract other namespace def
	quantify_tag(Tag,Default_ns,Local_ns,rdf:'Description'),			%quantify the tag, make sure it is 'rdf:Description'
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	%
	% zero or one idAboutAttr
	%
	parse_idaboutattr_zero_or_one(Attribs_left,New_default_ns,Subject,Attribs_temp1),		%get subject resource id/ref
	%
	% zero or more propAttr
	%
	parse_propattr_zero_or_more(Doc,Attribs_temp1,New_default_ns,Subject,_).			%get properties in attributes

%
% rdf:Description w/ nested element
%
parse_description(Doc,Default_ns,T,Subject) :-
	T=env(Tag,Attribs,Nested_list),						%element w/ nested elements
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),	%get any local namespace def
	extract_namespace_definition(Doc,Attribs_temp,Attribs_left),		%extract other namespace def
	quantify_tag(Tag,Default_ns,Local_ns,rdf:'Description'),		%quantify the tag, make sure it is 'rdf:Description'
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	%
	% get zero or one idAboutAttr
	%
	parse_idaboutattr_zero_or_one(Attribs_left,New_default_ns,Subject,Attribs_temp1),	%get subject resource id/ref
	%
	% get zero or more propAttr
	%
	parse_propattr_zero_or_more(Doc,Attribs_temp1,New_default_ns,Subject,[]),		%get properties in attributes
	%
	% parse zero or more property as element in value list
	%
	parse_propertyelt_zero_or_more(Doc,New_default_ns,Subject,Nested_list).		%parse nested elements

%
% abbreviated form w/ no nested element
%
parse_description(Doc,Default_ns,T,Subject) :-
	T=elem(Tag,Attribs),								%element w/ no nested element
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),		%get any local namespace def
	extract_namespace_definition(Doc,Attribs_temp,Attribs_left),			%extract other namespace def
	quantify_tag(Tag,Default_ns,Local_ns,Namespace:Name),				%quantify the tag
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	parse_idaboutattr_zero_or_one(Attribs_left,New_default_ns,Subject,Attribs_temp1),	%get subject id/about
	parse_propattr_zero_or_more(Doc,Attribs_temp1,New_default_ns,Subject,_),		%get properties from attributes
	%
	% subject's rdf:type is specified by tag
	%
	assert_rdf_type_if_needed(Doc,Subject,rdf:type,Namespace:Name).

%
% abbreviated form w/ nested element
%
parse_description(Doc,Default_ns,T,Subject) :-
	T=env(Tag,Attribs,Nested_list),							%element w/ nested elements
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),		%get any local namespace def
	extract_namespace_definition(Doc,Attribs_temp,Attribs_left),			%extract other namespace def
	quantify_tag(Tag,Default_ns,Local_ns,Namespace:Name),				%quantify the tag
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	parse_idaboutattr_zero_or_one(Attribs_left,New_default_ns,Subject,Attribs_temp1),	%get subject's id/about
	parse_propattr_zero_or_more(Doc,Attribs_temp1,New_default_ns,Subject,_),		%get properties from attributes
	%
	% parse zero or more property as element in value list
	%
	parse_propertyelt_zero_or_more(Doc,New_default_ns,Subject,Nested_list),
	%
	% subject's rdf:type is specified by tag
	%
	assert_rdf_type_if_needed(Doc,Subject,rdf:type,Namespace:Name).


%
% parse & extract zero or one idAboutAttrib, returns the resource ref as:
%
% +Doc		=	URL of the RDF document
% +Attribs	=	a list of attributes in the form X=Y
% +Default_ns	=	the default namespace
% -Id_or_about	=	the returned rdf ID or about by parsing the attribute list,
%			it is:
%			anonymous/1	-	no ID/about is found, make up a generic id
%			id/1		-	ID is found
% -Attribs_left	=	a list of remaining attributes
%
%
parse_idaboutattr_zero_or_one(Attribs,Default_ns,Id_or_about,Attribs_left) :-
	member(A,Attribs),						%find an element in the attributes
	parse_idaboutattr(A,Default_ns,Id_or_about),		%which is an ID/about
	!,
	delete(Attribs,A,Attribs_left).

%
% Originally generates "gensym_x".
% Now generates anonymous/1.
%
parse_idaboutattr_zero_or_one(Attribs,_,anonymous(Count),Attribs) :-
	generate_counter('anonymous_resource',Count,_).			%generate a generic resource name

%
% parse an rdf ID or about
%
% +Attrib	=	an attribute in the form X=Y
% +Namespace	=	the default namespace
% -Id_or_about	=	the returned ID or about
%
% [3] idAboutAttr    ::= idAttr | aboutAttr
%
parse_idaboutattr(A,Namespace,id(Id)) :-
	parse_idattr(A,Namespace,Id).

parse_idaboutattr(A,Namespace,About) :-
	parse_aboutattr(A,Namespace,About).

%
% parse an attribute which is 'rdf:about'
% if 'rdf:about' refers to a local ID, the term id(...) is returned
% if 'rdf:about' refers to a URI, the term uri(...) is returned
%
% +Attrib	=	an attrib in the form X=Y
% +Namespace	=	the default namespace
% -About	=	ID of the resource that it is refering to
%
% [4] aboutAttr      ::= 'about="' URI-reference '"'
%
parse_aboutattr(A=Value,Namespace,About) :-
	quantify_tag(A,Namespace,[],rdf:about),	%make sure that attribute is rdf:about
	parse_value(Value,Parsed_value),
	About=Parsed_value.
%	name(Value,Value_string),
%	(	(	Value_string=[0'#|Id_string],
%			!,
%			name(Id,Id_string),
%			About=id(Id)
%		);
%		About=uri(Value)
%	),
%	!.

%
% parse the 'rdf:ID' attribute, returns a id(...) term
%
% +Attrib	=	an attribute in the form X=Y
% +Namespace	=	the default namespace
% -Value	=	the parsed ID value as a id/1
%
%  [5] idAttr         ::= 'ID="' IDsymbol '"'
%
parse_idattr(A=Value,Namespace,Value) :-
	quantify_tag(A,Namespace,[],rdf:'ID').		%make sure that attribute is rdf:ID


%
% parse zero or more property in an attribute list
%
% +Doc			=	URL of the RDF document
% +Attribs		=	a list of attributes in the form X=Y
% +Namespace		=	the default namespace
% +Subject		=	resource ID of the subject
% -Attribs_remaining	=	the remaining list of attributes
%
parse_propattr_zero_or_more(_,[],_,_,[]).		%end of attribute list

parse_propattr_zero_or_more(Doc,[A|As],Namespace,Subject,Attribs_remaining) :-
	parse_propattr(Doc,A,Namespace,Subject),					%attribute parse as property
	parse_propattr_zero_or_more(Doc,As,Namespace,Subject,Attribs_remaining).

%
% parse an attribute which is property
%
% +Doc		=	URL of the RDF document
% +Attrib	=	an attribute in the form X=Y
% +Namespace	=	the default namespace
% +Subject	=	resource ID of the subject
%
parse_propattr(Doc,A=Value,Namespace,Subject) :-
	quantify_tag(A,Namespace,[],Quantified_property),			%find fully quantified property name
%
% parse value to see what it is
%
	parse_value(Value,Parsed_value),
	assertz(triple:rdf_content(Doc,triple(Subject,Quantified_property,Parsed_value))).			%assert triple

%
% parse value to see if it is a local ID reference,
% a URI or just a value
%
% +Value	=	the value of an attribute to parse
%			which is expected to be '#something', which 'something' is an ID
% -Id		=	the ID as an id/1 term
%
parse_value(Value,id(Id)) :-
	name(Value,Value_string),	%convert atomic value into a string
	Value_string=[0'#|String],	%remove '#' from the beginning
	!,
	name(Id,String).		%now get ID as an atom

%
% *** not parsing URI here
%
%parse_value(Value,Url) :-
%	pillow:url_info(Value,Url),
%	!.

parse_value(Value,Value).


%
% parse zero or more property element in a list
%
% +Doc		=	URL of the RDF document
% +Default_ns	=	the default namespace
% +Subject	=	the current subject under which these elements are properties
% +Elements	=	a list of elements
%
parse_propertyelt_zero_or_more(_,_,_,[]).

parse_propertyelt_zero_or_more(Doc,Default_ns,Subject,[E|Es]) :-
	parse_propertyelt(Doc,Default_ns,Subject,E),
	parse_propertyelt_zero_or_more(Doc,Default_ns,Subject,Es).

%
% parse a property as an element
%
%  [6a] propertyElt    ::= '<' propName resourceAttr? '/>'
%
% +Doc		=	URL of the RDF document
% +Default_ns	=	the default namespace
% +Subject	=	the subject under which this element is a property
% +Element	=	the element to parse
%
parse_propertyelt(_,_,_,E) :-
	E=comment(_),
	!.

parse_propertyelt(Doc,Default_ns,Subject,E) :-
	E=elem(Property_tag,Attribs),
	%
	% get any local namespace definition from attribute list
	%
	get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_temp),
	quantify_tag(Property_tag,Local_ns,Default_ns,Quantified_property),	%quantify tag
	Predicate=Quantified_property,
	parse_resource_attr_zero_or_one(Doc,Attribs_temp,Subject,Predicate,_).	%parse property value in attribute, which is a resource

%
%  [6a] propertyElt    ::= '<' propName '>' value '</' propName '>'
%
parse_propertyelt(Doc,Default_ns,Subject,E) :-
	E=env(Tag,Attribs,Value_list),
	quantify_tag(Tag,Local_ns,Default_ns,Quantified_property),		%quantify tag
	%
	% get any local namespace definition in attribute list
	%
	get_local_namespace_definition(Doc,Attribs,Local_ns,_Attribs_temp),
	%
	% find default namespace for the nested
	%
	first_non_empty([Local_ns,Default_ns],New_default_ns),
	%
	Predicate=Quantified_property,
	parse_property_value(Doc,New_default_ns,Subject,Predicate,Value_list).

%
% parse property value, which is either a literal or a single resource
%
% +Doc		=	URL of the RDF document
% +Default_ns	=	the default namespace
% +Subject	=	the current subject of the triple
% +Predicate	=	the predicate of the triple
% +Value	=	the value of this triple
%
parse_property_value(Doc,_,Subject,Predicate,[Value]) :-
	(	atom(Value);
		number(Value)
	),
	!,
	parse_value(Value,Parsed_value),
	Object=Parsed_value,
	assertz(triple:rdf_content(Doc,triple(Subject,Predicate,Object))).

parse_property_value(Doc,Default_ns,Subject,Predicate,[Value]) :-
	parse_description(Doc,Default_ns,Value,Object),
	assertz(triple:rdf_content(Doc,triple(Subject,Predicate,Object))).

%
% parse property value in a list of attributes as an "rdf:resource" attribute
%
% +Doc		=	URL of the RDF document
% +Attribs	=	a list of attributes in the form X=Y
% +Subject	=	the current subject of the triple
% +Predicate	=	the predicate of the triple of which we are looking for its value
% -Attribs_left	=	the remaining list of attributes
%
parse_resource_attr_zero_or_one(_,[],_,_,[]).				%end of attribute list

parse_resource_attr_zero_or_one(Doc,[A|As],Subject,Predicate,Attribs_left) :-
	parse_attrib_object(A,Subject,Predicate,Object),				%parse 1st attribute as an object
	!,
	parse_resource_attr_zero_or_one(Doc,As,Subject,Predicate,Attribs_left),		%parse other attributes
	assertz(triple:rdf_content(Doc,triple(Subject,Predicate,Object))).		%assert triple

parse_resource_attr_zero_or_one(Doc,[A|As],Subject,Predicate,[A|Attribs_left]) :-		%1st attribute not a rdf:resource
	parse_resource_attr_zero_or_one(Doc,As,Subject,Predicate,Attribs_left).		%parse other attributes

%
% parse the value of a single attribute "rdf:resource" as an object
%
% +Attrib	=	an attribute in the form X=Y
% +Subject	=	the subject of the triple
% +Predicate	=	the predicate of the triple
% -Parse_value	=	the parsed value
%
parse_attrib_object(Attrib=Value,_,Predicate,Parsed_value) :-
	Predicate=Property_ns:_,					%take namespace of property
	parse_value(Value,Parsed_value),
	quantify_tag(Attrib,Property_ns,[],rdf:resource).	%quantify tag, make sure it is rdf:resource



%
% split an atom by the colon, if any
%
% ?X	=	the atom to split
% ?Before	=	the part before the ':' as an atom
% ?After	=	the part after the ':' as an atom
%
split_by_colon(X,Before,After) :-
	nonvar(X),
	name(X,X_string),					%convert atom to string
	split_by_colon(Before_string,After_string,X_string,[]),
	split_by_colon_string_to_atom(Before_string,Before),
	split_by_colon_string_to_atom(After_string,After).

%
% merge 2 atoms into 1, separated by a colon
%
split_by_colon(X,Before,After) :-
	var(X),
	split_by_colon_string_to_atom(Before_string,Before),
	split_by_colon_string_to_atom(After_string,After),
	append(Before_string,":",Temp_string),
	append(Temp_string,After_string,X_string),
	name(X,X_string).

split_by_colon_string_to_atom([],[]) :-
	!.
split_by_colon_string_to_atom(String,Atom) :-
	name(Atom,String).


split_by_colon([B|Bs],[A|As]) -->
	alphabet(B),
	one_or_more_namespace_char(Bs),
	":",
	alphabet(A),
	one_or_more_namespace_char(As).

split_by_colon([],[A|As]) -->
	alphabet(A),
	one_or_more_namespace_char(As).

one_or_more_namespace_char([C|Cs]) -->
	namespace_char(C),
	one_or_more_namespace_char(Cs).

one_or_more_namespace_char([C]) -->
	namespace_char(C),
	{	!
	}.

alphabet(C) -->
	[C],
	{	nonvar(C),
		C >= 0'a,
		C =< 0'z
	}.

alphabet(C) -->
	[C],
	{	nonvar(C),
		C >= 0'A,
		C =< 0'Z
	}.

underscore(0'_) -->
	"_".

digit(C) -->
	[C],
	{	nonvar(C),
		C >=0'0,
		C =<0'9
	}.

%
% valid name-space character
%
namespace_char(C) -->
	alphabet(C).

namespace_char(C) -->
	digit(C).

namespace_char(C) -->
	underscore(C).

%
% reset a counter
%
reset_counter(Prefix) :-
	retractall(counter(Prefix,_)),
	asserta(counter(Prefix,1)).

%
% generate a counting value
%
generate_counter(Prefix,Count,Value) :-
	counter(Prefix,Count),
	!,
	Next_count is Count+1,
	retract(counter(Prefix,Count)),
	asserta(counter(Prefix,Next_count)),
	name(Prefix,Prefix_string),
	name(Count,Count_string),
	append(Prefix_string,Count_string,Value_string),
	name(Value,Value_string).

generate_counter(Prefix,Count,Value) :-
	Count=1,
	Next_count is Count+1,
	asserta(counter(Prefix,Next_count)),
	name(Prefix,Prefix_string),
	name(Count,Count_string),
	append(Prefix_string,Count_string,Value_string),
	name(Value,Value_string).


%
% find 1st non-empty element in a list
%
first_non_empty([],[]).
first_non_empty([[]|Xs],Non_empty) :-
	!,
	first_non_empty(Xs,Non_empty).
first_non_empty([X|_],X).


%
% given a tag (which may have a namespace specified),
% a default name-space, and a list of attributes which possibly define a local name-space
% fully quantify a tag bu finding out which name-space to use
%
quantify_tag(Tag,Default_ns,Local_ns,Quantified_tag) :-
%quantify_tag(Tag,Attribs,Default_ns,Local_namespace,Quantified_tag,Attribs_left) :-
	split_by_colon(Tag,Tag_ns,Name_only),					%split tag by colon, if any
	%
	% get for local namespace def in attributes, if any
	% var Local_ns will be '[]' if none is defined
	% otherwise it will be a namespace represented by a local/1 structure
	%
	%
	% find which namespace to use
	% 1st we consider the one in the tag,
	% then any local namespace defined in the attribute list
	% finally the default namespace
	%
	first_non_empty([Tag_ns,Local_ns,Default_ns],Namespace_to_use),	%select the one to use
	Quantified_tag=Namespace_to_use:Name_only.	%make up the fully-quantified tag

%
% get any local namespace def in an atttribute list
% if none is defined, return '[]'
%
get_local_namespace_definition(Doc,Attribs,Local_ns,Attribs_left) :-
	extract_local_namespace(Attribs,Local_list,Attribs_left),	%extract any local namespace definition from attributes
	valid_local_namespace_definition(Doc,Local_list,Local_ns).	%get the local namespace definition, if any

%
% extract_local_namespace/3
%
% extract all local namespace definition in a list of attributes,
% i.e. namespace definition with only 'xmlns', no name
%
extract_local_namespace([],[],[]).		%end of list

extract_local_namespace([xmlns=Uri|As],[Uri|Local],Remaining) :-		%a 'xmlns' definition
	!,
	extract_local_namespace(As,Local,Remaining).

extract_local_namespace([A|As],Local,[A|Remaining]) :-	%attribute not a 'xmlns'
	extract_local_namespace(As,Local,Remaining).		%process remaining list of attributes


%
% make sure that local name definition is either empty or single
% multiple definition is an error
%
valid_local_namespace_definition(_,[],[]) :-
	!.
valid_local_namespace_definition(Doc,[Uri],Local_ns) :-
	generate_counter(local_ns,Count,_),				%give a name for this locally defined ns
	Local_ns=local(Count),
	assertz(triple:rdf_content(Doc,namespace(Local_ns,Uri))),	%store local ns as namespace/2, name as local/1
	!.
%valid_local_namespace_definition(Tag,[_|_],_) :-
%	write('ERROR: multiple local namespace definition in <'),
%	write(Tag),
%	write('>'),
%	nl,
%	fail.


%
% extract XML name space definitions in an attribute list
% and store them as namespace/2
%
% +Doc	=	URL of this document
% +Attribs	=	a list of attributes in the form X=Y
% -Left		=	a list of remaining attributes which are not namespace definition
%
extract_namespace_definition(_,[],[]).

extract_namespace_definition(Doc,[A=V|As],Left) :-
	split_by_colon(A,xmlns,Name),					%split it by the colon ":"
	!,
	assertz(triple:rdf_content(Doc,namespace(Name,V))),		%assert it
	extract_namespace_definition(Doc,As,Left).

extract_namespace_definition(Doc,[A|As],[A|Left]) :-			%not a "xmlns:" something definition
	extract_namespace_definition(Doc,As,Left).


%
% assert rdf:type property triple if needed
% in the abbreviated form, this is implicit
% but if the property is not specified, then we add it
%
assert_rdf_type_if_needed(Doc,Subject,Predicate,Object) :-
	triple:rdf_content(Doc,triple(Subject,Predicate,Object)),
	!.

assert_rdf_type_if_needed(Doc,Subject,Predicate,Object) :-
	assertz(triple:rdf_content(Doc,triple(Subject,Predicate,Object))).


%--------------
% triple to RDF
%--------------

%
% starting from a resource
%
triple2rdf(Format,Rdf) :-
	make_rdf_header(Rdf),
	Rdf=env(_,_,Trees),
	all_resource_to_trees(Format,Trees).

triple2rdf(Format,Resource,Rdf) :-
	make_rdf_header(Rdf),
	Rdf=env(_,_,Tree),
	resource_to_tree(Format,Resource,[],Tree,_).

%
% makeup the RDF header w/ namespace definitions
%
make_rdf_header(Rdf_header) :-
	Rdf_header=env('rdf:RDF',Attribs,_),
	Template=(Namespace_def=Uri),
	Goal=(	namespace(Namespace,Uri),
		atom(Namespace),	%make sure that this is not a local/1 name
		concat_atoms(['xmlns:',Namespace],Namespace_def)
		),
	findall(Template,call(Goal),Attribs).
/*
all_resource_to_trees(Format,Trees) :-
	findall(Subject,triple(Subject,_,_),All_subjects),
	remove_duplicate(All_subjects,Subjects_no_dup),
	write('*** all subjects:'),
	write(Subjects_no_dup),
	nl,
	remove_object_resource(Subjects_no_dup,Remain),
	write('***remain:'),
	write(Remain),
	nl.
*/

remove_duplicate([],[]).
remove_duplicate([X|Xs],Result) :-
	remove_duplicate(Xs,No_dup),
	(	(	member(X,No_dup),
			Result=No_dup
		);
		Result=[X|No_dup]
	),
	!.

remove_can_reach([],[]).
remove_can_reach([S|Ss],Remain) :-
	remove_can_reach(S,Ss,Remain),
	remove_can_reach(Ss,Remain).

%remove_can_reach(S,Ss,Remain) :-
%	triple


remove_object_resource([],[]).

remove_object_resource([S|Ss],Remain) :-
	member(Someone,Ss),			%and is in the set Ss
	can_reach(Someone,S),			%Someone can reach S
	!,
	remove_object_resource(Ss,Remain).	%then S should be removed

remove_object_resource([S|Ss],[S|Remain]) :-
	remove_object_resource(Ss,Remain).

can_reach(X,Y) :-
	triple(X,_,Y).
can_reach(X,Y) :-
	triple(X,_,T),
	can_reach(T,Y).

/*
all_resource_to_trees(Format,Trees) :-
	%
	% find all subjects & their frequencies as object
	%
	Goal=(	triple(Subject,_,_),				%find an subject
		freq_as_object(Subject,Freq)			%and its frequency as subjects
		),
	Template=freq(Subject,Freq),
	findall(Template,call(Goal),Freq_list),			%find all subjects w/ their frequency as subjects
	%
	% find the one(s) w/ minimum frequency
	%
	find_min_freq(Freq_list,Min),
	write('*** frequency'),
	write(Min),
	nl,
	Min=freq(Resources,_),
	resource_to_tree(Format,Resources,[],Trees,_).
*/

freq_as_object(Something,Freq) :-
	findall(Something,triple(_,_,Something),List),
	length(List,Freq).

%	
% find item of a minimal frequencey in a list of freq/2
%	
find_min_freq([],[]).	%end of list, no min at all

find_min_freq([F|Fs],Min) :-
	find_min_freq(Fs,Min_temp),			%find min in remaining list
	find_min_freq_update(F,Min_temp,Min).		%see if min needs updating

%
% given a min, see if it needs updating
%
find_min_freq_update(F,[],F) :-						%no min, new one becomes min
	!.

find_min_freq_update(freq(X,Freq),freq(Xs,Freq),freq(Items,Freq)) :-	%same frequency, 1 more item, Xs is a list
	my_union(X,Xs,Items),						%take union to avoid duplicate
	!.

find_min_freq_update(freq(X,Freq),freq(_,Min_freq),freq(X,Freq)) :-
	Freq<Min_freq,							%this one is less frequent, replace old min
	!.

find_min_freq_update(_,Current,Current).				%this one is more frequent, keep old one



%
% converting a set of triple/3 into an XML(RDF) tree
% this doesn't include the <rdf:RDF> node
%

%
% handling a list of resources
%
resource_to_tree(_,[],Mapped,[],Mapped).

resource_to_tree(Format,[R|Rs],Mapped,[T|Ts],New_mapped) :-
	resource_to_tree(Format,R,Mapped,T,Mapped_temp),
	resource_to_tree(Format,Rs,Mapped_temp,Ts,New_mapped).




%
% resource is an atom or number, could be a URL as an atom
% this is a literal
%
resource_to_tree(_,Resource,Mapped,Resource,Mapped) :-
	literal(Resource),
	!.

%
% now resource is a term structure
% already mapped, make a rdf:about reference and no need to map subtree
%
resource_to_tree(Format,Resource,Mapped,Tree,New_mapped) :-
	member(Resource,Mapped),			%resource already mapped
	!,
	resource_to_tree_already_mapped(Format,Resource,Mapped,Tree,New_mapped).

%
% not yet mapped, have to map resource & its subtree
%
resource_to_tree(Format,Resource,Mapped,Tree,New_mapped) :-
	resource_to_tree_not_yet_mapped(Format,Resource,Mapped,Tree,New_mapped).

%
% resource already mapped is a id/1
% make a rdf:about attrib, no need to map the subtree
% abbreviated form
%
resource_to_tree_already_mapped(abbreviated,Resource,Mapped,Tree,Mapped) :-
	Resource=id(Id),
	triple(Resource,rdf:type,Namespace:Local),
	!,
	concat_atoms([Namespace,Local],Tag),
	concat_atoms(['#',Id],About),			%add the prefix hash
	Tree=elem(Tag,['rdf:about'=About]).		%make a rdf:about reference

%
% classic form
%
resource_to_tree_already_mapped(classic,Resource,Mapped,Tree,Mapped) :-
	Resource=id(Id),
	concat_atoms(['#',Id],About),			%add the prefix hash
	Tree=elem('rdf:Description',['about'=About]).

resource_to_tree_already_mapped(abbreviated,Resource,Mapped,_,Mapped) :-
	Resource=id(_),
	write('ERROR: resource rdf:type not found: '),
	write(Resource),
	nl,
	!,
	fail.






%
% resource already mapped is a anonymous/1
%
resource_to_tree_already_mapped(_,Resource,Mapped,_,Mapped) :-
	Resource=anonymous(_),
	write('ERROR: resource already mapped is an anonymous/1: '),
	write(Resource),
	nl,
	!,
	fail.

%
% resource already mapped is a namespace-quantified name
%
resource_to_tree_already_mapped(_,Resource,Mapped,_,Mapped) :-
	Resource=_:_,
	write('ERROR: resource already mapped is a namespace quantified name: '),
	write(Resource),
	nl,
	!,
	fail.

%
% resource not yet mapped
% this one is an id/1
%
resource_to_tree_not_yet_mapped(abbreviated,Resource,Mapped,Tree,New_mapped) :-
	Resource=id(Id),
	!,
	%
	% find all properties of this resource
	%
	Goal=triple(Resource,_,_),
	findall(Goal,call(Goal),Property_list),						%find properties as a list of triple/3
	find_subject_type(abbreviated,Resource,Property_list,Type,Property_left),	%find rdf:type, necessary for abbreviated form
	%
	% split properties into 2 lists
	%
	divide_properties(Property_left,Attrib_properties,Element_properties),
	%
	% compose resource node
	%
	Tree=env(Type,['rdf:ID'=Id|Attribs],Nested),
	property_to_attribs(Attrib_properties,Attribs),						%1 list of property become attributes
	property_to_tree(abbreviated,Element_properties,[Resource|Mapped],Nested,New_mapped).	%2nd one become nested elements

%
% classic form
%
resource_to_tree_not_yet_mapped(classic,Resource,Mapped,Tree,New_mapped) :-
	Resource=id(Id),
	!,
	Tree=env('rdf:Description',['ID'=Id],Nested),
	Goal=triple(Resource,_,_),
	findall(Goal,call(Goal),Property_list),						%find properties as a list of triple/3
	property_to_tree(classic,Property_list,[Resource|Mapped],Nested,New_mapped).	%2nd one become nested elements

%
% resource is a namespace quantified name
% abbreviated form
%
resource_to_tree_not_yet_mapped(abbreviated,Resource,Mapped,Uri,Mapped) :-
	Resource=_:_,
	!,
	quantified_namespace_to_uri(Resource,Uri).

%
% classic form
%
resource_to_tree_not_yet_mapped(classic,Resource,Mapped,Tree,Mapped) :-
	Resource=_:_,
	!,
	quantified_namespace_to_uri(Resource,Uri),
	Tree=elem('rdf:Description',['rdf:resource'=Uri]).

%
% resource is a anonymous/1
% abbreviated form
%
resource_to_tree_not_yet_mapped(abbreviated,Resource,Mapped,Tree,New_mapped) :-
	Resource=anonymous(_),
	!,
	%
	% find all properties of this resource
	%
	Goal=triple(Resource,_,_),
	findall(Goal,call(Goal),Property_list),
	find_subject_type(abbreviated,Resource,Property_list,Type,Property_left),	%find subject's type, necessary for abbreviated form
	%
	% split properties into 2 lists
	%
	divide_properties(Property_left,Attrib_properties,Element_properties),
	%
	% compose resource node
	%
	Tree=env(Type,Attribs,Nested),
	property_to_attribs(Attrib_properties,Attribs),						%1 list of property become attributes
	property_to_tree(abbreviated,Element_properties,[Resource|Mapped],Nested,New_mapped).	%2nd one become nested elements

%
% classic form
%
resource_to_tree_not_yet_mapped(classic,Resource,Mapped,Tree,New_mapped) :-
	Resource=anonymous(_),
	!,
	Tree=env('rdf:Description',[],Nested),
	Goal=triple(Resource,_,_),
	findall(Goal,call(Goal),Property_list),
	property_to_tree(classic,Property_list,[Resource|Mapped],Nested,New_mapped).	%2nd one become nested elements


%
% given a list of triple/3, find rdf:type in the list and remove it
% compose the tag name as a 'namespace:local_name' style atom
%
find_subject_type(abbreviated,Subject,Property_list,Type,Property_left) :-
	%
	% look for rdf:type
	%
	Template=triple(Subject,rdf:type,Namespace:Local),
	member(Template,Property_list),
	delete(Property_list,Template,Property_left),	%find & remove rdf:type from list
	concat_atoms([Namespace,':',Local],Type),
	!.

find_subject_type(abbreviated,Subject,_,_,_) :-
	write('ERROR: rdf:type not found for resource: '),
	write(Subject),
	nl,
	!,
	fail.

%
% map a list of property into XML trees
%
property_to_tree(_,[],Mapped,[],Mapped).

property_to_tree(abbreviated,[P|Ps],Mapped,[T|Ts],New_mapped) :-
	P=triple(_,Property,Object),				%the property triple to map
	Property=Namespace:Local,		%get namespace & local name
	concat_atoms([Namespace,':',Local],Tag),		%compose the property tag
	T=env(Tag,[],[Object_tree]),
	resource_to_tree(abbreviated,Object,Mapped,Object_tree,Mapped_temp),	%map object to XML tree (or atom if it is a literal)
	property_to_tree(abbreviated,Ps,Mapped_temp,Ts,New_mapped).		%process rest
%
% classic form
%

%
% object is a URL
%
property_to_tree(classic,[P|Ps],Mapped,[T|Ts],New_mapped) :-
	P=triple(_,Property,Object),				%the property triple to map
	Property=Namespace:Local,		%get namespace & local name
	parse_url:parse_url(Object,_),				%object is an URL
	concat_atoms([Namespace,':',Local],Tag),		%compose the property tag
	T=elem(Tag,['rdf:resource'=Object]),
	!,
	property_to_tree(classic,Ps,Mapped,Ts,New_mapped).		%process rest

property_to_tree(classic,[P|Ps],Mapped,[T|Ts],New_mapped) :-
	P=triple(_,Property,id(Id)),				%the property triple to map
	Property=Namespace:Local,		%get namespace & local name
	concat_atoms([Namespace,':',Local],Tag),		%compose the property tag
	concat_atoms(['#',Id],Id_ref),
	T=elem(Tag,['rdf:resource'=Id_ref]),
	!,
	property_to_tree(classic,Ps,Mapped,Ts,New_mapped).		%process rest

%
% object is another resource or a literal
%
property_to_tree(classic,[P|Ps],Mapped,[T|Ts],New_mapped) :-
	P=triple(_,Property,Object),				%the property triple to map
	Property=Namespace:Local,		%get namespace & local name
	concat_atoms([Namespace,':',Local],Tag),		%compose the property tag
	resource_to_tree(classic,Object,Mapped,Object_tree,Mapped_temp),	%map object to XML tree (or atom if it is a literal)
	(	(	literal(Object_tree),
			T=elem(Tag,['rdf:resource'=Object_tree])
		);
		T=env(Tag,[],[Object_tree])
	),
	!,
	property_to_tree(classic,Ps,Mapped_temp,Ts,New_mapped).		%process rest

%
% map a list of property into an attribute list
%
property_to_attribs([],[]).

property_to_attribs([P|Ps],[Tag=Value|As]) :-
	P=triple(_,Namespace:Local,Value),
	concat_atoms([Namespace,':',Local],Tag),
	property_to_attribs(Ps,As).


%
% map a namespace quantified name quantified name to a URI
%
quantified_namespace_to_uri(Namespace:Local_name,Uri) :-
	namespace(Namespace,Namespace_uri_with_hash),
	concat_atoms([Namespace_uri_with_hash,Local_name],Uri).

%
% given a list of triple/3, split it into 2 lists
% depending on the property should go into the attribute list or nested element tree
%
divide_properties([],[],[]).

divide_properties([P|Ps],[P|As],Es) :-
	property_as_attribute(List),
	P=triple(_,Pred,Object),
	member(Pred,List),			%this property goes into the attribute listS
	literal(Object),			%and object is a literal
	!,
	divide_properties(Ps,As,Es).

divide_properties([P|Ps],As,[P|Es]) :-
	divide_properties(Ps,As,Es).

%
% a literal is either an atom or a number, excluding an empty list
%
literal([]) :-
	!,
	fail.
literal(X) :-
	atom(X),
	!.
literal(X) :-
	number(X).

