% Eclipse socket library module
% Date : 2nd March
% Author : Stuart Chalmers
% adapted from Aberdeen socket library created by Kit Hui.

:- use_module(library(sockets)).			%use the ECLiPSe socket library

%---------------------------------------------
% create a socket in AF_INET or AF_UNIX domain
%
% a_socket_create(+Domain,-Socket)
%
% Domain=	'AF_INET' for Internet domain socket
%		'AF_UNIX' for UNIX domain socket
% Socket=	socket to be returned
%---------------------------------------------

a_socket_create('AF_INET',Socket) :-
	socket(internet,stream,Socket).

a_socket_create('AF_UNIX',Socket) :-
	socket(unix,stream,Socket).

%-----------------
% destroy a socket
% a_socket_destroy(+Socket)
%
% Socket=socket to be destroyed
%-----------------

a_socket_destroy(Socket):-
	close(Socket).

%-------------------------------
% binds a socket with an address
% a_socket_bind(+Socket,+Address)
%
% Socket=	socket to be binded (created by a_socket_create/2)
% Address=	address in the form 'AF_INET'(port)
% or		'AF_UNIX'(pathname)
%-------------------------------

a_socket_bind(Socket,'AF_INET'(Port)):-
	current_host(Host),	
	bind(Socket,Host/Port).

		
%----------------------------------------
% listen to a socket and set queue length
% a_socket_listen(+Socket,+Length)
%
% Socket=	socket to listen (created by a_socket_create/2)
% Length=	integer specifying the queue length
%----------------------------------------

a_socket_listen(Socket,Queue_length):-
	listen(Socket,Queue_length).

%------------------------------
% accept an incoming connection
% a_socket_accept(+Socket,-Income,-Read,Write)
%
% Socket=	socket for accepting incoming connection
% Income=	returned file-descriptor of incoming connection
% Read=		read stream for incoming connection
% Write=	write stream for incoming connection
%------------------------------

a_socket_accept(Socket,New_stream,New_stream,New_stream):-
	accept(Socket,_,New_stream).


%---------------------------
% connect to a remote socket
% a_socket_connect(+Socket,+Address,+Read,+Write)
%
% Socket=	socket for making connection (created by a_socket_create/2)
% Address=	target address for connection in the form 'AF_INET'(host,port)
% or		'AF_UNIX'(pathname) for the UNIX domain
% Read=		read stream to be returned for connection
% Write=	write stream to be returned
%---------------------------

a_socket_connect(Socket,'AF_INET'(Host, Port),Socket,Socket):-
	connect(Socket,Host/Port).

%------------------------------
% get the current hostname
% a_hostname(-Host)
%
% Host=	hostname to be returned
%------------------------------

a_hostname(Host_Name):-
	get_flag(hostname,Host),
	term_string(Host_Name,Host).


%------------------------------------------------------------------------
%
% check to see if socket or streams are ready for reading
%
%
% a_socket_select(+Sockets,+Read_streams,+Timeout,-Ready_connections)
%
% Sockets=	a list of sockets
% Read_streams=	a list of reading streams
% Timeout=	timeout value in the form time(second,micro-second)
%		a value of 0 assume blocking
% Reading_connections=	a list of connections which are ready for reading
%			in the form connection(socket,read,write)
%
%------------------------------------------------------------------------

%
% trap error when there is more than 1 socket
%
a_socket_select([_,_|_],_,_,_) :-
	write('*** a_socket_select/4 supports one socket only in ECLiPSe!!!***'),
	nl,
	!,
	fail.

%
% no socket
%
a_socket_select([],Read_streams,0,Ready_connections) :-
	stream_select(Read_streams,off,Ready_read_streams),			%check for new connection & ready read streams
	findall(connection(_,R,R),member(R,Ready_read_streams),Ready_connections).

a_socket_select([],Read_streams,time(Sec,Usec),Ready_connections) :-
	(	(	Usec=<0,						%ECLiPSe doesn'tlike Usec =<0
			!,
			Msec=1
		);
		Msec=Usec
	),
	stream_select(Read_streams,Sec:Msec,Ready_read_streams),		%check for new connection & ready read streams
	findall(connection(_,R,R),member(R,Ready_read_streams),Ready_connections).

%
% exactly 1 socket
%
a_socket_select([Socket],Read_streams,0,Ready_connections) :-
	socket_select(Socket,New_from_socket,off,Read_streams,Ready_read_streams),	%check for new connection & ready read streams
	findall(connection(_,R,R),member(R,Ready_read_streams),Conn_from_read),
	(	(	nonvar(New_from_socket),					%if a connection is got from socket
			!,
			append([connection(_,New_from_socket,New_from_socket)],Conn_from_read,Ready_connections)
		);
		Ready_connections=Conn_from_read
	).

a_socket_select([Socket],Read_streams,time(Sec,Usec),Ready_connections) :-
	(	(	Usec=<0,						%ECLiPSe doesn'tlike Usec =<0
			!,
			Msec=1
		);
		Msec=Usec
	),
	socket_select(Socket,New_from_socket,Sec:Msec,Read_streams,Ready_read_streams),	%check for new connection & ready read streams
	findall(connection(_,R,R),member(R,Ready_read_streams),Conn_from_read),
	(	(	nonvar(New_from_socket),					%if a connection is got from socket
			!,
			append([connection(_,New_from_socket,New_from_socket)],Conn_from_read,Ready_connections)
		);
		Ready_connections=Conn_from_read
	).

/*
a_socket_select(Sockets,Read_streams,0,Ready_connections) :-
	select(Read_streams,block,From_read),					%check read streams
        a_socket_select(Sockets,off,Conn_from_sockets),				%check sockets
	findall(connection(_,Y,Y),member(Y,From_read),Conn_from_streams),	%get list of connection/3 from read streams
	append(Conn_from_sockets,Conn_from_streams,Ready_connections).

a_socket_select(Sockets,Read_streams,time(Sec,_),Ready_connections) :-
	select(Read_streams,Sec,From_read),					%check read streams
        a_socket_select(Sockets,Sec,Conn_from_sockets),				%check sockets
        findall(connection(_,Y,Y),member(Y,From_read),Conn_from_streams),	%get list of connection/3 from read streams
        append(Conn_from_sockets,Conn_from_streams,Ready_connections).
*/

%
% a_socket_select/3
%
% ECLiPSe socket_select/5 takes a single socket instead of a list
% so we have to iterate through the list
%
a_socket_select([],_,[]).
a_socket_select([S|Ss],Timeout,[C|Cs]) :-
        socket_select(S,New_stream,Timeout,[],_),	%see if socket is ready
	C=connection(S,New_stream,New_stream),		%make up connection/3 term
	a_socket_select(Ss,Timeout,Cs).			%check the rest in list


%----------------------------------------------------------------------
% select from lists of file descriptors
% a_socket_select(+Reads,+Writes,+Excepts,+Timeout,-Read,-Write,-Except)
%
% Reads=	list of file-descriptors for testing read-ready
% Writes=	list of file-descriptors for testing write-ready
% Excepts=	list of file-descriptors for testing exceptions
% Timeout=	time-out value for selection
%		either 0 for no timeout (blocking)
%		or time(Second,Micro-second) for timeout
% Read=		returned list of file-descriptors ready for reading
% Write=	returned list of file-descriptors ready for writing
% Except=	returned list of file-descriptors with exception
%----------------------------------------------------------------------


% if socket is the empty list then don't do select - returns an error
a_socket_select(Rs,Ws,Es,time(Secs,MSecs),R,W,E):-
	(MSecs > 500,
	Time is Secs + 1;
	Time is Secs),
	(Rs = [],R=[];select(Rs,Time,R)),
	(Ws = [],W=[];select(Ws,Time,W)),
	(Es = [],E=[];select(Es,Time,E)),!.
		
a_socket_select(Rs,Ws,Es,0,R,W,E):-
	(Rs = [],R=[];select(Rs,block,R)),
	(Ws = [],W=[];select(Ws,block,W)),
	(Es = [],E=[];select(Es,block,E)),!.


	

