%
% socket library for SWI Prolog
%

:- use_module(library(socket)).		%use SWI C-lib's socket library **notice: no 's' after 'socket'


%-------------------------------------------------------------------------------
% a_socket_create(+Domain,-Socket)
%
% create a socket
%
% Domain=	socket domain. One of AF_INET or AF_UNIX
% Socket=	socket created
%-------------------------------------------------------------------------------

a_socket_create('AF_UNIX',_)	:-
	write('*** AF_UNIX domain not supported by SWI'),
	nl,
	!,
	fail.

a_socket_create('AF_INET',Socket)	:-
	!,
	var(Socket),
	socket:tcp_socket(Socket).



%-------------------------------------------------------------------------------
% a_socket_destroy(+Socket)
%
% destroy a socket
%
% Socket=	socket to be destroyed
%-------------------------------------------------------------------------------

a_socket_destroy(Socket)	:-
	nonvar(Socket),
	socket:tcp_close_socket(Socket).




%-------------------------------------------------------------------------------
% a_socket_listen(+Socket,+Length)
%
% listen to a socket and set queue length
%
% Socket=	Socket to listen to
% Length=	length of queue
%-------------------------------------------------------------------------------

a_socket_listen(Socket,Length)	:-
	socket:tcp_listen(Socket,Length).


%-------------------------------------------------------------------------------
% a_socket_bind(+Socket,+Address)
%
% binds a socket to a name
%
% Socket=	socket to be binded
% Address=	address to bind. One of the forms:
%			'AF_INET'(Host,Port)
%		or	'AF_UNIX'(Pathname)
%-------------------------------------------------------------------------------

a_socket_bind(Socket,'AF_INET'(Port))	:-
	!,
	socket:tcp_bind(Socket,Port).

a_socket_bind(Socket,Port)	:-
	socket:tcp_bind(Socket,Port).

%-------------------------------------------------------------------------------
% a_socket_connect(+Socket,+Address,-Read,-Write)
%
% connect a socket to a remote address
%
% Socket=	socket for making connection
% Address=	remote address to connect to. It is in the form:
%			'AF_INET'(Host:Port)
%		or	'AF_UNIX'(Pathname)
% Read=		read stream returned
% Write=	write stream returned
%-------------------------------------------------------------------------------

a_socket_connect(Socket,'AF_INET'(Host,Port),Read,Write)	:-
	nonvar(Host),
	nonvar(Port),
	var(Read),
	var(Write),
	!,
	socket:tcp_connect(Socket,Host:Port),
	socket:tcp_open_socket(Socket,Read,Write).



%-------------------------------------------------------------------------------
% a_socket_accept(+Socket,-Income,-Read,-Write)
%
% accept an incoming connection
%
% Socket=	socket to accept from
% Income=	income socket accepted
% Read=		read-stream returned
% Write=	write-stream returned
%-------------------------------------------------------------------------------

a_socket_accept(Socket,Income,Read,Write)	:-
	nonvar(Socket),
	!,
	socket:tcp_accept(Socket,Income,_),
	socket:tcp_open_socket(Income,Read,Write).


%-----------------------------------------------
% a_socket_select(+Sockets,+Reads,+Timeout,-Ready)
%
% check if a list of sockets or read streams are ready for reading
%
% Sockets=	a list of sockets
% Reads=	list of read streams
% Timeout=	time-out value for testing
%		0 for blocking selection
%		time(sec,u-sec) for time-out
% Ready=	the list of ready connection
%
% NOTE:	The way SWI test for readiness in socket & stream is by the wait_for_input/3 predicate,
%	which takes only a list of streams but NO socket.
%	To check if there is a connection to a server socket, we have to open a stream on it
%	using tcp_open_socket/3, then include the opened stream (on the server socket) in calling
%	wait_for_input/3.
%	Once a stream is returned by wait_for_input/3, we have to see if it comes from a server socket.
%	If yes, we have to accept the incoming connection (by tcp_accept/3) and then open streams
%	from it (using tcp_open_stream/3).
%
%-----------------------------------------------

a_socket_select(Sockets,Reads,0,Connections) :-
	open_stream_from_server_socket(Sockets,Streams_from_sockets,Socket_stream_pairs),	%open streams from server sockets
	append(Streams_from_sockets,Reads,Streams_to_test),					%collect all streams in a list
	wait_for_input(Streams_to_test,Ready_streams,0),					%check for readiness
	find_connection(Ready_streams,Socket_stream_pairs,Connections).

a_socket_select(Sockets,Reads,time(Sec,Usec),Connections) :-
	Timeout is Sec+Usec/1000,
	open_stream_from_server_socket(Sockets,Streams_from_sockets,Socket_stream_pairs),	%open streams from server sockets
	append(Streams_from_sockets,Reads,Streams_to_test),					%collect all streams in a list
	wait_for_input(Streams_to_test,Ready_streams,Timeout),					%check for readiness
	find_connection(Ready_streams,Socket_stream_pairs,Connections).


%
% give a list of sockets, open streams for them
%
open_stream_from_server_socket([],[],[]).
open_stream_from_server_socket([S|Ss],[X|Xs],[socket_stream(S,X)|Ps]) :-
	socket:tcp_open_socket(S,X,_),
	open_stream_from_server_socket(Ss,Xs,Ps).

find_connection([],_,[]).
find_connection([R|Rs],Socket_stream_pairs,[C|Cs]) :-
	member(socket_stream(Socket,R),Socket_stream_pairs),	%stream is from a server socket
	!,
	tcp_accept(Socket,Incoming,_),				%accept from this server socket
	tcp_open_socket(Incoming,New_read,New_write),		%open streams from it
	C=connection(Incoming,New_read,New_write),
	find_connection(Rs,Socket_stream_pairs,Cs).
find_connection([R|Rs],Socket_stream_pairs,[C|Cs]) :-
	C=connection(_,R,_),
	find_connection(Rs,Socket_stream_pairs,Cs).




%------------------------------------------------------------------------------
% a_socket_select(+Reads,+Writes,+Excepts,+Timeout,-R_ready,-W_ready,-E_ready)
%
% select from a list of file-descriptors
%
% Reads=	list of file-descriptors for read-ready test
% Writes=	list of file-descriptors for write-ready test
% Excepts=	list of file-descriptors for exception test
% Timeout=	time-out value for testing
%		0 for blocking selection
%		time(sec,u-sec) for time-out
% R_ready=	returned list of read-ready fd
% W_ready=	returned list of write-ready fd
% E_ready=	returned list of fds with exception
%------------------------------------------------------------------------------


%
%
%
a_socket_select(R,W,E,0,RR,WR,ER)	:-
	socket_select1(R,W,E,0,0,RR,WR,ER).

a_socket_select(R,W,E,time(Sec,Usec),RR,WR,ER)	:-
	socket_select1(R,W,E,Sec,Usec,RR,WR,ER).



%----------------------------------
% get the fully quantified hostname
%----------------------------------

a_hostname(Host)	:-
	socket:gethostname(Host).			%get host name
%	socket:gethostname(Host),			%get host name
%	socket:tcp_host_to_address(Host,ip(A,B,C,D)),	%convert host to IP

